# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT PACKAGES AND MODULES
# ///////////////////////////////////////////////////////////////
from gui.widgets.py_table_widget.py_table_widget import PyTableWidget
from . functions_main_window import *
import sys
import os

# IMPORT QT CORE
# ///////////////////////////////////////////////////////////////
from qt_core import *

# IMPORT SETTINGS
# ///////////////////////////////////////////////////////////////
from gui.core.json_settings import Settings

# IMPORT THEME COLORS
# ///////////////////////////////////////////////////////////////
from gui.core.json_themes import Themes

# IMPORT PY ONE DARK WIDGETS
# ///////////////////////////////////////////////////////////////
from gui.widgets import *

# LOAD UI MAIN
# ///////////////////////////////////////////////////////////////
from . ui_main import *

# MAIN FUNCTIONS 
# ///////////////////////////////////////////////////////////////
from . functions_main_window import *

# PY WINDOW
# ///////////////////////////////////////////////////////////////
class SetupMainWindow:
    def __init__(self):
        super().__init__()
        # SETUP MAIN WINDOw
        # Load widgets from "gui\uis\main_window\ui_main.py"
        # ///////////////////////////////////////////////////////////////
        self.ui = UI_MainWindow()
        self.ui.setup_ui(self)

    # ADD LEFT MENUS
    # ///////////////////////////////////////////////////////////////
    add_left_menus = [
        {
            "btn_icon" : "icon_home.svg",
            "btn_id" : "btn_home",
            "btn_text" : "Home",
            "btn_tooltip" : "Home page",
            "show_top" : True,
            "is_active" : False
        },
        {
            "btn_icon" : "icon_widgets.svg",
            "btn_id" : "btn_widgets",
            "btn_text" : "Show Custom Widgets",
            "btn_tooltip" : "Show custom widgets",
            "show_top" : True,
            "is_active" : False
        },
        {
            "btn_icon" : "icon_add_user.svg",
            "btn_id" : "btn_add_user",
            "btn_text" : "Add Users",
            "btn_tooltip" : "Add users",
            "show_top" : True,
            "is_active" : False
        },
        {
            "btn_icon" : "icon_scope.svg",
            "btn_id" : "btn_scope",
            "btn_text" : "Scope",
            "btn_tooltip" : "View scope",
            "show_top" : True,
            "is_active" : True # 记得修改下面的初始页面
        },
        # {
        #     "btn_icon" : "icon_file.svg",
        #     "btn_id" : "btn_new_file",
        #     "btn_text" : "New File",
        #     "btn_tooltip" : "Create new file",
        #     "show_top" : True,
        #     "is_active" : False
        # },
        # {
        #     "btn_icon" : "icon_folder_open.svg",
        #     "btn_id" : "btn_open_file",
        #     "btn_text" : "Open File",
        #     "btn_tooltip" : "Open file",
        #     "show_top" : True,
        #     "is_active" : False
        # },
        # {
        #     "btn_icon" : "icon_save.svg",
        #     "btn_id" : "btn_save",
        #     "btn_text" : "Save File",
        #     "btn_tooltip" : "Save file",
        #     "show_top" : True,
        #     "is_active" : False
        # },
        {
            "btn_icon" : "icon_info.svg",
            "btn_id" : "btn_info",
            "btn_text" : "Information",
            "btn_tooltip" : "Open informations",
            "show_top" : False,
            "is_active" : False
        },
        {
            "btn_icon" : "icon_settings.svg",
            "btn_id" : "btn_settings",
            "btn_text" : "Settings",
            "btn_tooltip" : "Open settings",
            "show_top" : False,
            "is_active" : False
        }
    ]

     # ADD TITLE BAR MENUS
    # ///////////////////////////////////////////////////////////////
    add_title_bar_menus = [
        {
            "btn_icon" : "film-svgrepo-com.svg",
            "btn_id" : "btn_run_realtime_simulation",
            "btn_tooltip" : "Run RealTime-ASMSimPy",
            "is_active" : False
        },
        # {
        #     "btn_icon" : "icon_search.svg",
        #     "btn_id" : "btn_search",
        #     "btn_tooltip" : "Search",
        #     "is_active" : False
        # },
        {
            "btn_icon" : "icon_settings.svg",
            "btn_id" : "btn_top_settings",
            "btn_tooltip" : "Top settings",
            "is_active" : False
        }
    ]

    # SETUP CUSTOM BTNs OF CUSTOM WIDGETS
    # Get sender() function when btn is clicked
    # ///////////////////////////////////////////////////////////////
    def setup_btns(self):
        if self.ui.title_bar.sender() != None:
            return self.ui.title_bar.sender()
        elif self.ui.left_menu.sender() != None:
            return self.ui.left_menu.sender()
        elif self.ui.left_column.sender() != None:
            return self.ui.left_column.sender()

    # SETUP MAIN WINDOW WITH CUSTOM PARAMETERS
    # ///////////////////////////////////////////////////////////////
    def setup_gui(self):
        # APP TITLE
        # ///////////////////////////////////////////////////////////////
        self.setWindowTitle(self.settings["app_name"])
        
        # REMOVE TITLE BAR
        # ///////////////////////////////////////////////////////////////
        if self.settings["custom_title_bar"]:
            self.setWindowFlag(Qt.FramelessWindowHint)
            self.setAttribute(Qt.WA_TranslucentBackground)

        # ADD GRIPS
        # ///////////////////////////////////////////////////////////////
        if self.settings["custom_title_bar"]:
            self.left_grip = PyGrips(self, "left", self.hide_grips)
            self.right_grip = PyGrips(self, "right", self.hide_grips)
            self.top_grip = PyGrips(self, "top", self.hide_grips)
            self.bottom_grip = PyGrips(self, "bottom", self.hide_grips)
            self.top_left_grip = PyGrips(self, "top_left", self.hide_grips)
            self.top_right_grip = PyGrips(self, "top_right", self.hide_grips)
            self.bottom_left_grip = PyGrips(self, "bottom_left", self.hide_grips)
            self.bottom_right_grip = PyGrips(self, "bottom_right", self.hide_grips)

        # LEFT MENUS / GET SIGNALS WHEN LEFT MENU BTN IS CLICKED / RELEASED
        # ///////////////////////////////////////////////////////////////
        # ADD MENUS
        self.ui.left_menu.add_menus(SetupMainWindow.add_left_menus)

        # SET SIGNALS
        self.ui.left_menu.clicked.connect(self.btn_clicked)
        self.ui.left_menu.released.connect(self.btn_released)

        # TITLE BAR / ADD EXTRA BUTTONS
        # ///////////////////////////////////////////////////////////////
        # ADD MENUS
        self.ui.title_bar.add_menus(SetupMainWindow.add_title_bar_menus)

        # SET SIGNALS
        self.ui.title_bar.clicked.connect(self.btn_clicked)
        self.ui.title_bar.released.connect(self.btn_released)

        # ADD Title
        if self.settings["custom_title_bar"]:
            self.ui.title_bar.set_title(self.settings["app_name"])
        else:
            self.ui.title_bar.set_title("Welcome to PyOneDark")

        # LEFT COLUMN SET SIGNALS
        # ///////////////////////////////////////////////////////////////
        self.ui.left_column.clicked.connect(self.btn_clicked)
        self.ui.left_column.released.connect(self.btn_released)

        # SET INITIAL PAGE / SET LEFT AND RIGHT COLUMN MENUS
        # ///////////////////////////////////////////////////////////////
        MainFunctions.set_page(self, self.ui.load_pages.page_4_scope) # 选择初始页面为page_4_scope
        MainFunctions.set_left_column_menu(
            self,
            menu = self.ui.left_column.menus.menu_1,
            title = "Settings Left Column",
            icon_path = Functions.set_svg_icon("icon_settings.svg")
        )
        MainFunctions.set_right_column_menu(self, self.ui.right_column.menu_1)

        # ///////////////////////////////////////////////////////////////
        # EXAMPLE CUSTOM WIDGETS
        # Here are added the custom widgets to pages and columns that
        # were created using Qt Designer.
        # This is just an example and should be deleted when creating
        # your application.
        #
        # OBJECTS FOR LOAD PAGES, LEFT AND RIGHT COLUMNS
        # You can access objects inside Qt Designer projects using
        # the objects below:
        #
        # <OBJECTS>
        # LEFT COLUMN: self.ui.left_column.menus
        # RIGHT COLUMN: self.ui.right_column
        # LOAD PAGES: self.ui.load_pages
        # </OBJECTS>
        # ///////////////////////////////////////////////////////////////

        # LOAD SETTINGS
        # ///////////////////////////////////////////////////////////////
        settings = Settings()
        self.settings = settings.items

        # LOAD THEME COLOR
        # ///////////////////////////////////////////////////////////////
        themes = Themes()
        self.themes = themes.items

        # LEFT COLUMN
        # ///////////////////////////////////////////////////////////////

        # BTN 1
        self.left_btn_1 = PyPushButton(
            text="Btn 1",
            radius=8,
            color=self.themes["app_color"]["text_foreground"],
            bg_color=self.themes["app_color"]["dark_one"],
            bg_color_hover=self.themes["app_color"]["dark_three"],
            bg_color_pressed=self.themes["app_color"]["dark_four"]
        )
        self.left_btn_1.setMaximumHeight(40)
        self.ui.left_column.menus.btn_1_layout.addWidget(self.left_btn_1)

        # BTN 2
        self.left_btn_2 = PyPushButton(
            text="Btn With Icon",
            radius=8,
            color=self.themes["app_color"]["text_foreground"],
            bg_color=self.themes["app_color"]["dark_one"],
            bg_color_hover=self.themes["app_color"]["dark_three"],
            bg_color_pressed=self.themes["app_color"]["dark_four"]
        )
        self.icon = QIcon(Functions.set_svg_icon("icon_settings.svg"))
        self.left_btn_2.setIcon(self.icon)
        self.left_btn_2.setMaximumHeight(40)
        self.ui.left_column.menus.btn_2_layout.addWidget(self.left_btn_2)

        # BTN 3 - Default QPushButton
        self.left_btn_3 = QPushButton("Default QPushButton")
        self.left_btn_3.setMaximumHeight(40)
        self.ui.left_column.menus.btn_3_layout.addWidget(self.left_btn_3)

        # PAGES
        # ///////////////////////////////////////////////////////////////

        # PAGE 1 - ADD LOGO TO MAIN PAGE
        self.logo_svg = QSvgWidget(Functions.set_svg_image("logo_home.svg"))
        self.ui.load_pages.logo_layout.addWidget(self.logo_svg, Qt.AlignCenter, Qt.AlignCenter)

        # PAGE 2
        # CIRCULAR PROGRESS 1
        self.circular_progress_1 = PyCircularProgress(
            value = 80,
            progress_color = self.themes["app_color"]["context_color"],
            text_color = self.themes["app_color"]["text_title"],
            font_size = 14,
            bg_color = self.themes["app_color"]["dark_four"]
        )
        self.circular_progress_1.setFixedSize(200,200)

        # CIRCULAR PROGRESS 2
        self.circular_progress_2 = PyCircularProgress(
            value = 45,
            progress_width = 4,
            progress_color = self.themes["app_color"]["context_color"],
            text_color = self.themes["app_color"]["context_color"],
            font_size = 14,
            bg_color = self.themes["app_color"]["bg_three"]
        )
        self.circular_progress_2.setFixedSize(160,160)

        # CIRCULAR PROGRESS 3
        self.circular_progress_3 = PyCircularProgress(
            value = 75,
            progress_width = 2,
            progress_color = self.themes["app_color"]["pink"],
            text_color = self.themes["app_color"]["white"],
            font_size = 14,
            bg_color = self.themes["app_color"]["bg_three"]
        )
        self.circular_progress_3.setFixedSize(140,140)

        # PY SLIDER 1
        self.vertical_slider_1 = PySlider(
            margin=8,
            bg_size=10,
            bg_radius=5,
            handle_margin=-3,
            handle_size=16,
            handle_radius=8,
            bg_color = self.themes["app_color"]["dark_three"],
            bg_color_hover = self.themes["app_color"]["dark_four"],
            handle_color = self.themes["app_color"]["context_color"],
            handle_color_hover = self.themes["app_color"]["context_hover"],
            handle_color_pressed = self.themes["app_color"]["context_pressed"]
        )
        self.vertical_slider_1.setMinimumHeight(100)

        # PY SLIDER 2
        self.vertical_slider_2 = PySlider(
            bg_color = self.themes["app_color"]["dark_three"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            handle_color = self.themes["app_color"]["context_color"],
            handle_color_hover = self.themes["app_color"]["context_hover"],
            handle_color_pressed = self.themes["app_color"]["context_pressed"]
        )
        self.vertical_slider_2.setMinimumHeight(100)

        # PY SLIDER 3
        self.vertical_slider_3 = PySlider(
            margin=8,
            bg_size=10,
            bg_radius=5,
            handle_margin=-3,
            handle_size=16,
            handle_radius=8,
            bg_color = self.themes["app_color"]["dark_three"],
            bg_color_hover = self.themes["app_color"]["dark_four"],
            handle_color = self.themes["app_color"]["context_color"],
            handle_color_hover = self.themes["app_color"]["context_hover"],
            handle_color_pressed = self.themes["app_color"]["context_pressed"]
        )
        self.vertical_slider_3.setOrientation(Qt.Horizontal)
        self.vertical_slider_3.setMaximumWidth(200)

        # PY SLIDER 4
        self.vertical_slider_4 = PySlider(
            bg_color = self.themes["app_color"]["dark_three"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            handle_color = self.themes["app_color"]["context_color"],
            handle_color_hover = self.themes["app_color"]["context_hover"],
            handle_color_pressed = self.themes["app_color"]["context_pressed"]
        )
        self.vertical_slider_4.setOrientation(Qt.Horizontal)
        self.vertical_slider_4.setMaximumWidth(200)

        # ICON BUTTON 1
        self.icon_button_1 = PyIconButton(
            icon_path = Functions.set_svg_icon("icon_heart.svg"),
            parent = self,
            app_parent = self.ui.central_widget,
            tooltip_text = "Icon button - Heart",
            width = 40,
            height = 40,
            radius = 20,
            dark_one = self.themes["app_color"]["dark_one"],
            icon_color = self.themes["app_color"]["icon_color"],
            icon_color_hover = self.themes["app_color"]["icon_hover"],
            icon_color_pressed = self.themes["app_color"]["icon_active"],
            icon_color_active = self.themes["app_color"]["icon_active"],
            bg_color = self.themes["app_color"]["dark_one"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            bg_color_pressed = self.themes["app_color"]["pink"]
        )

        # ICON BUTTON 2
        self.icon_button_2 = PyIconButton(
            icon_path = Functions.set_svg_icon("icon_add_user.svg"),
            parent = self,
            app_parent = self.ui.central_widget,
            tooltip_text = "BTN with tooltip",
            width = 40,
            height = 40,
            radius = 8,
            dark_one = self.themes["app_color"]["dark_one"],
            icon_color = self.themes["app_color"]["icon_color"],
            icon_color_hover = self.themes["app_color"]["icon_hover"],
            icon_color_pressed = self.themes["app_color"]["white"],
            icon_color_active = self.themes["app_color"]["icon_active"],
            bg_color = self.themes["app_color"]["dark_one"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            bg_color_pressed = self.themes["app_color"]["green"],
        )

        # ICON BUTTON 3
        self.icon_button_3 = PyIconButton(
            icon_path = Functions.set_svg_icon("icon_add_user.svg"),
            parent = self,
            app_parent = self.ui.central_widget,
            tooltip_text = "BTN actived! (is_actived = True)",
            width = 40,
            height = 40,
            radius = 8,
            dark_one = self.themes["app_color"]["dark_one"],
            icon_color = self.themes["app_color"]["icon_color"],
            icon_color_hover = self.themes["app_color"]["icon_hover"],
            icon_color_pressed = self.themes["app_color"]["white"],
            icon_color_active = self.themes["app_color"]["icon_active"],
            bg_color = self.themes["app_color"]["dark_one"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            bg_color_pressed = self.themes["app_color"]["context_color"],
            is_active = True
        )

        # PUSH BUTTON 1
        self.push_button_1 = PyPushButton(
            text = "Button Without Icon",
            radius  =8,
            color = self.themes["app_color"]["text_foreground"],
            bg_color = self.themes["app_color"]["dark_one"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            bg_color_pressed = self.themes["app_color"]["dark_four"]
        )
        self.push_button_1.setMinimumHeight(40)

        # PUSH BUTTON 2
        self.push_button_2 = PyPushButton(
            text = "Button With Icon",
            radius = 8,
            color = self.themes["app_color"]["text_foreground"],
            bg_color = self.themes["app_color"]["dark_one"],
            bg_color_hover = self.themes["app_color"]["dark_three"],
            bg_color_pressed = self.themes["app_color"]["dark_four"]
        )
        self.icon_2 = QIcon(Functions.set_svg_icon("icon_settings.svg"))
        self.push_button_2.setMinimumHeight(40)
        self.push_button_2.setIcon(self.icon_2)

        # PY LINE EDIT
        self.line_edit = PyLineEdit(
            text = "",
            place_holder_text = "Place holder text",
            radius = 8,
            border_size = 2,
            color = self.themes["app_color"]["text_foreground"],
            selection_color = self.themes["app_color"]["white"],
            bg_color = self.themes["app_color"]["dark_one"],
            bg_color_active = self.themes["app_color"]["dark_three"],
            context_color = self.themes["app_color"]["context_color"]
        )
        self.line_edit.setMinimumHeight(30)

        # TOGGLE BUTTON
        self.toggle_button = PyToggle(
            width = 50,
            bg_color = self.themes["app_color"]["dark_two"],
            circle_color = self.themes["app_color"]["icon_color"],
            active_color = self.themes["app_color"]["context_color"]
        )

        # TABLE WIDGETS
        self.table_widget = PyTableWidget(
            radius = 8,
            color = self.themes["app_color"]["text_foreground"],
            selection_color = self.themes["app_color"]["context_color"],
            bg_color = self.themes["app_color"]["bg_two"],
            header_horizontal_color = self.themes["app_color"]["dark_two"],
            header_vertical_color = self.themes["app_color"]["bg_three"],
            bottom_line_color = self.themes["app_color"]["bg_three"],
            grid_line_color = self.themes["app_color"]["bg_one"],
            scroll_bar_bg_color = self.themes["app_color"]["bg_one"],
            scroll_bar_btn_color = self.themes["app_color"]["dark_four"],
            context_color = self.themes["app_color"]["context_color"]
        )
        self.table_widget.setColumnCount(3)
        self.table_widget.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.table_widget.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.table_widget.setSelectionBehavior(QAbstractItemView.SelectRows)

        # Columns / Header
        self.column_1 = QTableWidgetItem()
        self.column_1.setTextAlignment(Qt.AlignCenter)
        self.column_1.setText("NAME")

        self.column_2 = QTableWidgetItem()
        self.column_2.setTextAlignment(Qt.AlignCenter)
        self.column_2.setText("NICK")

        self.column_3 = QTableWidgetItem()
        self.column_3.setTextAlignment(Qt.AlignCenter)
        self.column_3.setText("PASS")

        # Set column
        self.table_widget.setHorizontalHeaderItem(0, self.column_1)
        self.table_widget.setHorizontalHeaderItem(1, self.column_2)
        self.table_widget.setHorizontalHeaderItem(2, self.column_3)

        for x in range(10):
            row_number = self.table_widget.rowCount()
            self.table_widget.insertRow(row_number) # Insert row
            self.table_widget.setItem(row_number, 0, QTableWidgetItem(str("Wanderson"))) # Add name
            self.table_widget.setItem(row_number, 1, QTableWidgetItem(str("vfx_on_fire_" + str(x)))) # Add nick
            self.pass_text = QTableWidgetItem()
            self.pass_text.setTextAlignment(Qt.AlignCenter)
            self.pass_text.setText("12345" + str(x))
            self.table_widget.setItem(row_number, 2, self.pass_text) # Add pass
            self.table_widget.setRowHeight(row_number, 22)

        # ADD WIDGETS
        self.ui.load_pages.row_1_layout.addWidget(self.circular_progress_1)
        self.ui.load_pages.row_1_layout.addWidget(self.circular_progress_2)
        self.ui.load_pages.row_1_layout.addWidget(self.circular_progress_3)
        self.ui.load_pages.row_2_layout.addWidget(self.vertical_slider_1)
        self.ui.load_pages.row_2_layout.addWidget(self.vertical_slider_2)
        self.ui.load_pages.row_2_layout.addWidget(self.vertical_slider_3)
        self.ui.load_pages.row_2_layout.addWidget(self.vertical_slider_4)
        self.ui.load_pages.row_3_layout.addWidget(self.icon_button_1)
        self.ui.load_pages.row_3_layout.addWidget(self.icon_button_2)
        self.ui.load_pages.row_3_layout.addWidget(self.icon_button_3)
        self.ui.load_pages.row_3_layout.addWidget(self.push_button_1)
        self.ui.load_pages.row_3_layout.addWidget(self.push_button_2)
        self.ui.load_pages.row_3_layout.addWidget(self.toggle_button)
        self.ui.load_pages.row_4_layout.addWidget(self.line_edit)
        self.ui.load_pages.row_5_layout.addWidget(self.table_widget)

        # RIGHT COLUMN
        # ///////////////////////////////////////////////////////////////

        # BTN 1
        self.right_btn_1 = PyPushButton(
            text="Show Menu 2",
            radius=8,
            color=self.themes["app_color"]["text_foreground"],
            bg_color=self.themes["app_color"]["dark_one"],
            bg_color_hover=self.themes["app_color"]["dark_three"],
            bg_color_pressed=self.themes["app_color"]["dark_four"]
        )
        self.icon_right = QIcon(Functions.set_svg_icon("icon_arrow_right.svg"))
        self.right_btn_1.setIcon(self.icon_right)
        self.right_btn_1.setMaximumHeight(40)
        self.right_btn_1.clicked.connect(lambda: MainFunctions.set_right_column_menu(
            self,
            self.ui.right_column.menu_2
        ))
        self.ui.right_column.btn_1_layout.addWidget(self.right_btn_1)

        # BTN 2
        self.right_btn_2 = PyPushButton(
            text="Show Menu 1",
            radius=8,
            color=self.themes["app_color"]["text_foreground"],
            bg_color=self.themes["app_color"]["dark_one"],
            bg_color_hover=self.themes["app_color"]["dark_three"],
            bg_color_pressed=self.themes["app_color"]["dark_four"]
        )
        self.icon_left = QIcon(Functions.set_svg_icon("icon_arrow_left.svg"))
        self.right_btn_2.setIcon(self.icon_left)
        self.right_btn_2.setMaximumHeight(40)
        self.right_btn_2.clicked.connect(lambda: MainFunctions.set_right_column_menu(
            self,
            self.ui.right_column.menu_1
        ))
        self.ui.right_column.btn_2_layout.addWidget(self.right_btn_2)

        # ///////////////////////////////////////////////////////////////
        # END - EXAMPLE CUSTOM WIDGETS
        # ///////////////////////////////////////////////////////////////

    # RESIZE GRIPS AND CHANGE POSITION
    # Resize or change position when window is resized
    # ///////////////////////////////////////////////////////////////
    def resize_grips(self):
        if self.settings["custom_title_bar"]:
            self.left_grip.setGeometry(5, 10, 10, self.height())
            self.right_grip.setGeometry(self.width() - 15, 10, 10, self.height())
            self.top_grip.setGeometry(5, 5, self.width() - 10, 10)
            self.bottom_grip.setGeometry(5, self.height() - 15, self.width() - 10, 10)
            self.top_right_grip.setGeometry(self.width() - 20, 5, 15, 15)
            self.bottom_left_grip.setGeometry(5, self.height() - 20, 15, 15)
            self.bottom_right_grip.setGeometry(self.width() - 20, self.height() - 20, 15, 15)



# %%
try:
    from numba.experimental import jitclass
except:
    from numba import jitclass
from numba import njit#, types, vectorize, prange
from numba import int32, float64 # import the types
from pylab import np, plt
plt.style.use('ggplot')


############################################# CLASS DEFINITION 
@jitclass(
    spec=[
        # CONSOLE
            # ('reset', int32),
            ('sensorless_position', int32),
            ('sensorless_speed', int32),
            ('CMD_SPEED_SINE_HZ', float64),
            ('CMD_SPEED_SINE_RPM', float64),
        # CONTROL
            ('use_disturbance_feedforward_rejection', int32),
            ('KP', float64),
        # MOTOR
            # name plate data
            ('init_npp',   int32),
            ('init_IN',  float64),
            # electrical parameters
            ('init_R',   float64),
            ('init_Ld',  float64),
            ('init_Lq',  float64),
            ('init_KE',  float64),
            ('init_Rreq',float64),
            # mechanical parameters
            ('init_Js',  float64),
    ])
class The_Human:
    def __init__(self):
        # self.reset = False
        self.sensorless_position = False
        self.sensorless_speed    = False
        self.CMD_SPEED_SINE_HZ   = 10
        self.CMD_SPEED_SINE_RPM  = 0
        # ADRC
        self.use_disturbance_feedforward_rejection = 0
        self.KP = 100
        # name plate data
        self.init_npp = 4
        self.init_IN = 3 # Arms (line-to-line)
        # electrical parameters
        self.init_R = 1.1
        self.init_Ld = 5e-3
        self.init_Lq = 6e-3 # 5e-3 #BUG
        self.init_KE = 0.095
        self.init_Rreq = -1.0 # this should be set to infinity but setting to a negative value is easier for coding
        # mechanical parameters
        self.init_Js = 0.0006168  # kg.m^2

@jitclass(
    spec=[
        # name plate data
        ('npp',   int32),
        ('npp_inv', float64),
        ('IN',  float64),
        # electrical parameters
        ('R',   float64),
        ('Ld',  float64),
        ('Lq',  float64),
        ('KE',  float64),
        ('Rreq',float64),
        # mechanical parameters
        ('Js',  float64),
        ('Js_inv', float64),
        # states
        ('NS',    int32),
        ('x',   float64[:]),
        ('Tem', float64),
        # inputs
        ('uab',   float64[:]),
        ('udq',   float64[:]),
        ('TLoad', float64),
        # output
        ('cosT', float64),
        ('sinT', float64),
        ('iab',  float64[:]),
        ('idq',  float64[:]),
        ('omega_elec', float64),
        ('theta_d', float64),
        ('theta_mech', float64),
        ('KActive', float64),
    ])
class The_AC_Machines:
    def __init__(self, HUMAN):
        # name plate data
        self.npp = HUMAN.init_npp
        self.npp_inv = 1.0/self.npp
        self.IN  = HUMAN.init_IN
        # electrical parameters
        self.R   = HUMAN.init_R
        self.Ld  = HUMAN.init_Ld
        self.Lq  = HUMAN.init_Lq
        self.KE  = HUMAN.init_KE
        self.Rreq  = HUMAN.init_Rreq
        # mechanical parameters
        self.Js  = HUMAN.init_Js # kg.m^2
        self.Js_inv = 1.0/self.Js
        # states
        self.NS = 6
        self.x = np.zeros(self.NS, dtype=np.float64)
        self.x[5] = self.KE
        self.Tem = 0.0
        # inputs
        self.uab = np.zeros(2, dtype=np.float64)
        self.udq = np.zeros(2, dtype=np.float64)
        self.TLoad = 0
        # output
        self.cosT = 0.0
        self.sinT = 0.0
        self.iab = np.zeros(2, dtype=np.float64)
        self.idq = np.zeros(2, dtype=np.float64)
        self.omega_elec = 0.0
        self.theta_d = 0.0
        self.theta_mech = 0.0
        self.KActive = self.KE

@jitclass(
    spec=[
        # constants
        ('CL_TS', float64),
        ('VL_TS', float64),
        # feedback / input
        ('theta_d', float64),
        ('omega_elec', float64),
        ('iab', float64[:]),
        # states
        ('timebase', float64),
        ('cosT', float64),
        ('sinT', float64),
        ('idq', float64[:]),
        ('KActive', float64),
        ('Tem', float64),
        # commands
        ('cmd_idq', float64[:]),
        ('cmd_udq', float64[:]),
        ('cmd_uab', float64[:]),
        ('cmd_rpm', float64),
        # MOTOR
            # name plate data
            ('npp',   int32),
            ('IN',  float64),
            # electrical parameters
            ('R',   float64),
            ('Ld',  float64),
            ('Lq',  float64),
            ('KE',  float64),
            ('Rreq',float64),
            # mechanical parameters
            ('Js',  float64),
    ])
class The_Motor_Controller:
    def __init__(self, CL_TS, VL_TS, HUMAN):
        # constants
        self.CL_TS = CL_TS
        self.VL_TS = VL_TS
        # feedback / input
        self.theta_d = 0.0
        self.omega_elec = 0.0
        self.iab = np.zeros(2, dtype=np.float64)
        # states
        self.timebase = 0.0
        self.cosT = 0.0
        self.sinT = 0.0
        self.idq = np.zeros(2, dtype=np.float64)
        self.KActive = 0.0
        self.Tem = 0.0
        # commands 
        self.cmd_idq = np.zeros(2, dtype=np.float64)
        self.cmd_udq = np.zeros(2, dtype=np.float64)
        self.cmd_uab = np.zeros(2, dtype=np.float64)
        self.cmd_rpm = 0.0
        # MOTOR
        self.npp = HUMAN.init_npp
        self.IN  = HUMAN.init_IN
        self.R   = HUMAN.init_R
        self.Ld  = HUMAN.init_Ld
        self.Lq  = HUMAN.init_Lq
        self.KE  = HUMAN.init_KE
        self.Rreq = HUMAN.init_Rreq
        self.Js  = HUMAN.init_Js

@jitclass(
    spec=[
        ('Kp', float64),
        ('Ki', float64),
        ('Err', float64),
        ('Ref', float64),
        ('Fbk', float64),
        ('Out', float64),
        ('OutLimit', float64),
        ('ErrPrev', float64),
        ('OutPrev', float64),
    ])
class The_PI_Regulator:
    def __init__(self, KP_CODE, KI_CODE, OUTPUT_LIMIT):
        self.Kp = KP_CODE
        self.Ki = KI_CODE
        self.Err      = 0.0
        self.Ref      = 0.0
        self.Fbk      = 0.0
        self.Out      = 0.0
        self.OutLimit = OUTPUT_LIMIT
        self.ErrPrev  = 0.0
        self.OutPrev  = 0.0

############################################# OBSERVER SECTION

@jitclass(
    spec=[
        # feedback / inputs
        ('uab', float64[:]),
        ('uab_prev', float64[:]),
        ('uab_curr', float64[:]),
        ('iab', float64[:]),
        ('iab_prev', float64[:]),
        ('iab_curr', float64[:]),
        # states
        ('NS', int32),
        ('NS_SPEED', int32),
        ('NS_FLUX', int32),
        ('xspeed', float64[:]),
        ('xflux', float64[:]),
        ('cosT', float64),
        ('sinT', float64),
        # outputs
        ('theta_d', float64),
        ('omega_elec', float64),
        ('total_disrubance_feedforward', float64),
        ('output_error', float64),
        ('amplitude_mismatch', float64[:]),
        ('stator_flux', float64[:]),
        ('u_offset', float64[:]),
        ('active_flux', float64[:]),
        ('active_flux_ampl', float64),
        ('active_flux_ampl_lpf', float64),
        ('ActiveFlux_KP', float64),
        ('ActiveFlux_KI', float64),
        # gains
        ('ell1', float64),
        ('ell2', float64),
        ('ell3', float64),
        ('ell4', float64),
        ('one_over_six', float64),
        # MOTOR
            # # name plate data
            # ('npp',   int32),
            # ('IN',  float64),
            # # electrical parameters
            # ('R',   float64),
            # ('Ld',  float64),
            # ('Lq',  float64),
            # ('KE',  float64),
            # ('Rreq',float64),
            # # mechanical parameters
            # ('Js',  float64),
    ])
class The_Observer:
    def __init__(self, HUMAN):
        # feedback / input
        self.uab      = np.zeros(2, dtype=np.float64)
        self.uab_prev = np.zeros(2, dtype=np.float64)
        self.uab_curr = np.zeros(2, dtype=np.float64)
        self.iab      = np.zeros(2, dtype=np.float64)
        self.iab_prev = np.zeros(2, dtype=np.float64)
        self.iab_curr = np.zeros(2, dtype=np.float64)
        # state
        self.NS       = 4 # = max(NS_SPEED, NS_FLUX)
        self.NS_SPEED = 4
        self.NS_FLUX  = 4
        self.xspeed = np.zeros(self.NS_SPEED, dtype=np.float64)
        self.xflux = np.zeros(self.NS_FLUX, dtype=np.float64)
        self.cosT = 1.0
        self.sinT = 0.0
        # outputs
        self.theta_d    = 0.0
        self.omega_elec = 0.0
        self.total_disrubance_feedforward = 0.0
        self.output_error = 0.0
        self.amplitude_mismatch = np.zeros(2, dtype=np.float64)
        self.stator_flux = np.zeros(2, dtype=np.float64)
        self.u_offset = np.zeros(2, dtype=np.float64)
        self.active_flux = np.zeros(2, dtype=np.float64)
        self.active_flux[0] = HUMAN.init_KE
        self.active_flux_ampl = HUMAN.init_KE
        self.active_flux_ampl_lpf = HUMAN.init_KE
        self.ActiveFlux_KP = 200
        self.ActiveFlux_KI = 0.0

        # gains
        omega_ob = 100 # [rad/s]
        if False: # 2nd-order speed observer (assuming speed feedback)
            self.ell2 = 2 * omega_ob
            self.ell3 =     omega_ob**2 * HUMAN.init_Js/HUMAN.init_npp
        elif True: # 3rd-order position observer
            self.ell1 = 3 * omega_ob
            self.ell2 = 3 * omega_ob**2
            self.ell3 =     omega_ob**3 * HUMAN.init_Js/HUMAN.init_npp
            self.ell4 = 0.0
        else: # 4th-order position observer
            self.ell1 = 4 * omega_ob
            self.ell2 = 6 * omega_ob**2
            self.ell3 = 4 * omega_ob**3 * HUMAN.init_Js/HUMAN.init_npp
            self.ell4 =     omega_ob**4

        self.one_over_six = 1.0 / 6.0

        # MOTOR
        # self.npp = HUMAN.init_npp
        # self.IN  = HUMAN.init_IN
        # self.R   = HUMAN.init_R
        # self.Ld  = HUMAN.init_Ld
        # self.Lq  = HUMAN.init_Lq
        # self.KE  = HUMAN.init_KE
        # self.Rreq = HUMAN.init_Rreq
        # self.Js  = HUMAN.init_Js

@njit(nogil=True)
def DYNAMICS_2ndOdESO_LOF(x, OB, CTRL):
    fx = np.zeros(OB.NS_SPEED)

    # [rad/s]
    output_error = CTRL.omega_elec - x[1]

    # 机械子系统 (omega_r_elec, theta_d, theta_r_mech)
    fx[0] = x[1]
    fx[1] = OB.ell2*output_error + (CTRL.Tem + x[2]) * CTRL.npp/CTRL.Js # elec. angular rotor speed
    fx[2] = OB.ell3*output_error + 0.0
    fx[3] = 0.0
    return fx

@njit(nogil=True)
def DYNAMICS_4thOdESO_LOF(x, OB, CTRL):
    fx = np.zeros(OB.NS_SPEED)

    # [rad]
    output_error = np.sin(CTRL.theta_d - x[0])

    # 机械子系统 (omega_r_elec, theta_d, theta_r_mech)
    fx[0] = OB.ell1*output_error + x[1]
    fx[1] = OB.ell2*output_error + (CTRL.Tem + x[2]) * CTRL.npp/CTRL.Js # elec. angular rotor speed
    fx[2] = OB.ell3*output_error + x[3]
    fx[3] = OB.ell4*output_error + 0.0
    return fx

@njit(nogil=True)
def DYNAMICS_AFE_25_VMCMFusion(x, OB, CTRL):
    fx = np.zeros(OB.NS_FLUX)

    OB.active_flux[0] = x[0] - CTRL.Lq * CTRL.iab[0]
    OB.active_flux[1] = x[1] - CTRL.Lq * CTRL.iab[1]
    OB.active_flux_ampl = np.sqrt(OB.active_flux[0]*OB.active_flux[0] + OB.active_flux[1]*OB.active_flux[1])

    if OB.active_flux_ampl!=0:
        active_flux_ampl_inv = 1.0/OB.active_flux_ampl
        OB.cosT = OB.active_flux[0] * active_flux_ampl_inv
        OB.sinT = OB.active_flux[1] * active_flux_ampl_inv
    else:
        OB.cosT = 1.0
        OB.sinT = 0.0
    
    iD_at_current_step = CTRL.iab[0] * OB.cosT + CTRL.iab[1] * OB.sinT
    # iQ_at_current_step = CTRL.iab[0] *-OB.sinT + CTRL.iab[1] * OB.cosT
    CTRL.KActive = (CTRL.Ld - CTRL.Lq) * iD_at_current_step + CTRL.KE # 有功磁链计算

    OB.amplitude_mismatch[0] = CTRL.KActive*OB.cosT - OB.active_flux[0]
    OB.amplitude_mismatch[1] = CTRL.KActive*OB.sinT - OB.active_flux[1]

    fx[0] = CTRL.cmd_uab[0] - CTRL.R*CTRL.iab[0] \
            + OB.ActiveFlux_KP * OB.amplitude_mismatch[0] \
            + x[2]
    fx[1] = CTRL.cmd_uab[1] - CTRL.R*CTRL.iab[1] \
            + OB.ActiveFlux_KP * OB.amplitude_mismatch[1] \
            + x[3]

    fx[2] = OB.ActiveFlux_KI * OB.amplitude_mismatch[0]
    fx[3] = OB.ActiveFlux_KI * OB.amplitude_mismatch[1]
    return fx

@njit(nogil=True)
def RK4_OBSERVER_CJH_STYLE(THE_DYNAMICS, x, OB, hs, CTRL):
    NS = OB.NS # THIS SHOULD BE A CONSTANT THROUGHOUT THE CODES!!!
    k1, k2, k3, k4 = np.zeros(NS), np.zeros(NS), np.zeros(NS), np.zeros(NS) # incrementals at 4 stages
    xk, fx = np.zeros(NS), np.zeros(NS) # state x for stage 2/3/4, state derivative

    OB.uab[0] = OB.uab_prev[0]
    OB.uab[1] = OB.uab_prev[1]
    OB.iab[0] = OB.iab_prev[0]
    OB.iab[1] = OB.iab_prev[1]
    fx = THE_DYNAMICS(x, OB, CTRL)
    for i in range(0, NS):
        k1[i] = fx[i] * hs
        xk[i] = x[i] + k1[i]*0.5
                                                                         
    OB.iab[0] = 0.5*(OB.iab_prev[0]+OB.iab_curr[0])
    OB.iab[1] = 0.5*(OB.iab_prev[1]+OB.iab_curr[1])
    OB.uab[0] = 0.5*(OB.uab_prev[0]+OB.uab_curr[0])
    OB.uab[1] = 0.5*(OB.uab_prev[1]+OB.uab_curr[1])
    fx = THE_DYNAMICS(xk, OB, CTRL)
    for i in range(0, NS):
        k2[i] = fx[i] * hs
        xk[i] = x[i] + k2[i]*0.5
                                                                         
    fx = THE_DYNAMICS(xk, OB, CTRL)
    for i in range(0, NS):
        k3[i] = fx[i] * hs
        xk[i] = x[i] + k3[i]
                                                                         
    OB.iab[0] = OB.iab_curr[0]
    OB.iab[1] = OB.iab_curr[1]
    OB.uab[0] = OB.uab_curr[0]
    OB.uab[1] = OB.uab_curr[1]
    fx = THE_DYNAMICS(xk, OB, CTRL)
    for i in range(0, NS):
        k4[i] = fx[i] * hs
        x[i] = x[i] + (k1[i] + 2*(k2[i] + k3[i]) + k4[i]) * OB.one_over_six

############################################# MACHINE SIMULATION SECTION

@njit(nogil=True)
def DYNAMICS_MACHINE(t, x, ACM, CLARKE_TRANS_TORQUE_GAIN=1.5):
    fx = np.zeros(ACM.NS)

    # 电磁子系统 (id, iq, KActive as x[0], x[1], x[5])
    if ACM.Rreq > 0:
    # IM: Rreq is positive
        fx[5] = ACM.Rreq*x[0] - (ACM.Ld - ACM.Lq)/ACM.Rreq * x[5]
        ACM.KActive = x[5]
        fx[0] = (ACM.udq[0] - ACM.R*x[0] + x[2]*ACM.Lq*x[1] + fx[5]) / ACM.Lq
    else:
    # PMSM: Rreq is infinitely large
        # ACM.KActive = (ACM.Ld - ACM.Lq) * x[0] + ACM.KE # version 1
        ACM.KActive = x[5]                              # version 2 
        fx[0] = (ACM.udq[0] - ACM.R*x[0] + x[2]*ACM.Lq*x[1]) / ACM.Ld
        fx[5] = (ACM.Ld - ACM.Lq) * fx[0] + 0.0

    fx[1] = (ACM.udq[1] - ACM.R*x[1] - x[2]*ACM.Lq*x[0] - x[2]*ACM.KActive) / ACM.Lq

    # 机械子系统 (omega_r_elec, theta_d, theta_r_mech as x[2], x[3], x[4])
    ACM.Tem = CLARKE_TRANS_TORQUE_GAIN * ACM.npp * ACM.KActive *x[1] # 电磁转矩计算
    fx[2] = (ACM.Tem - ACM.TLoad) * ACM.npp/ACM.Js # elec. angular rotor speed
    fx[3] = x[2]          # elec. angular rotor position (bounded)
    fx[4] = x[2]/ACM.npp  # mech. angular rotor position (accumulated)
    return fx

@njit(nogil=True)
def RK4_MACHINE(t, ACM, hs): # 四阶龙格库塔法
    NS = ACM.NS
    k1, k2, k3, k4 = np.zeros(NS), np.zeros(NS), np.zeros(NS), np.zeros(NS) # incrementals at 4 stages
    xk, fx = np.zeros(NS), np.zeros(NS) # state x for stage 2/3/4, state derivative

    if False:
        """ this is about twice slower than loop through the element one by one """ 
        fx = DYNAMICS_MACHINE(t, ACM.x, ACM) # @t
        k1 = fx * hs
        xk = ACM.x + k1*0.5

        fx = DYNAMICS_MACHINE(t, xk, ACM)  # @t+hs/2
        k2 = fx * hs
        xk = ACM.x + k2*0.5

        fx = DYNAMICS_MACHINE(t, xk, ACM)  # @t+hs/2
        k3 = fx * hs
        xk = ACM.x + k3

        fx = DYNAMICS_MACHINE(t, xk, ACM)  # @t+hs
        k4 = fx * hs
        ACM.x = ACM.x + (k1 + 2*(k2 + k3) + k4)/6.0
    else:
        fx = DYNAMICS_MACHINE(t, ACM.x, ACM) # @t
        for i in range(NS):
            k1[i] = fx[i] * hs
            xk[i] = ACM.x[i] + k1[i]*0.5

        fx = DYNAMICS_MACHINE(t, xk, ACM)  # @t+hs/2
        for i in range(NS):
            k2[i] = fx[i] * hs
            xk[i] = ACM.x[i] + k2[i]*0.5

        fx = DYNAMICS_MACHINE(t, xk, ACM)  # @t+hs/2
        for i in range(NS):
            k3[i] = fx[i] * hs
            xk[i] = ACM.x[i] + k3[i]

        fx = DYNAMICS_MACHINE(t, xk, ACM)  # @t+hs
        for i in range(NS):
            k4[i] = fx[i] * hs
            ACM.x[i] = ACM.x[i] + (k1[i] + 2*(k2[i] + k3[i]) + k4[i])/6.0
            # ACM.x_dot[i] = (k1[i] + 2*(k2[i] + k3[i]) + k4[i])/6.0 / hs # derivatives

############################################# BASIC FOC SECTION

@njit(nogil=True)
def incremental_pi(reg):
    reg.Err = reg.Ref - reg.Fbk
    reg.Out = reg.OutPrev + \
        reg.Kp * (reg.Err - reg.ErrPrev) + \
        reg.Ki * reg.Err
    if reg.Out >    reg.OutLimit:
        reg.Out =   reg.OutLimit
    elif reg.Out < -reg.OutLimit:
        reg.Out =  -reg.OutLimit
    reg.ErrPrev = reg.Err
    reg.OutPrev = reg.Out

@njit(nogil=True)
def FOC(CTRL, reg_id, reg_iq, reg_speed, OB, HUMAN):
    reg_speed.Ref = CTRL.cmd_rpm / 60 * 2*np.pi * CTRL.npp
    reg_speed.Fbk = CTRL.omega_elec
    incremental_pi(reg_speed)

    reg_id.Ref = CTRL.cmd_idq[0]
    reg_id.Fbk = CTRL.idq[0]
    incremental_pi(reg_id)
    CTRL.cmd_udq[0] = reg_id.Out

    if HUMAN.use_disturbance_feedforward_rejection == 0:
        CTRL.cmd_idq[1] = reg_speed.Out
    else:
        CTRL.cmd_idq[1] = HUMAN.KP*(reg_speed.Ref-reg_speed.Fbk) + OB.total_disrubance_feedforward

    reg_iq.Ref = CTRL.cmd_idq[1]
    reg_iq.Fbk = CTRL.idq[1]
    incremental_pi(reg_iq)
    CTRL.cmd_udq[1] = reg_iq.Out

    return CTRL.cmd_udq

############################################# DSP SECTION

""" DSP """
@njit(nogil=True)
def DSP(ACM, CTRL, reg_id, reg_iq, reg_speed, OB, HUMAN):
    CTRL.timebase += CTRL.CL_TS

    """ Current Measurment """
    CTRL.iab[0] = ACM.iab[0] # + offset, scale error , noise
    CTRL.iab[1] = ACM.iab[1] # + offset, scale error , noise

    """ Position Measurement """
    if not HUMAN.sensorless_position:
        CTRL.theta_d = ACM.theta_d + 0.10*0 #BUG???

    """ Position and Speed Estimator """
    RK4_OBSERVER_CJH_STYLE(DYNAMICS_AFE_25_VMCMFusion, OB.xflux, OB, CTRL.CL_TS, CTRL)
    # Unpack x
    OB.stator_flux[0] = OB.xflux[0]
    OB.stator_flux[1] = OB.xflux[1]
    OB.u_offset[0] = OB.xflux[2]
    OB.u_offset[1] = OB.xflux[3]
    # Active flux
    OB.active_flux[0] = OB.xflux[0] - CTRL.Lq * CTRL.iab[0]
    OB.active_flux[1] = OB.xflux[1] - CTRL.Lq * CTRL.iab[1]
    OB.theta_d = np.arctan2(OB.active_flux[1], OB.active_flux[0])
    OB.active_flux_ampl = np.sqrt(OB.active_flux[0]*OB.active_flux[0] + OB.active_flux[1]*OB.active_flux[1])
    OB.active_flux_ampl_lpf = OB.active_flux_ampl #_lpf(OB.active_flux_ampl, OB.active_flux_ampl_lpf, 5)
    if OB.active_flux_ampl!=0:
        active_flux_ampl_inv = 1.0/OB.active_flux_ampl
        OB.cosT = OB.active_flux[0] * active_flux_ampl_inv
        OB.sinT = OB.active_flux[1] * active_flux_ampl_inv
    else:
        OB.cosT = 1.0
        OB.sinT = 0.0
    OB.amplitude_mismatch[0] = CTRL.KActive*OB.cosT - OB.active_flux[0]
    OB.amplitude_mismatch[1] = CTRL.KActive*OB.sinT - OB.active_flux[1]

    if HUMAN.sensorless_position:
        CTRL.theta_d = OB.theta_d

    """ Park Transformation Essentials
    """
    # do this once per control interrupt
    CTRL.cosT = np.cos(CTRL.theta_d)
    CTRL.sinT = np.sin(CTRL.theta_d)
    # Park transformation
    CTRL.idq[0] = CTRL.iab[0] * CTRL.cosT + CTRL.iab[1] * CTRL.sinT
    CTRL.idq[1] = CTRL.iab[0] *-CTRL.sinT + CTRL.iab[1] * CTRL.cosT

    """ Speed Observer """
    CTRL.KActive = (CTRL.Ld - CTRL.Lq) * CTRL.idq[0] + CTRL.KE # 有功磁链计算
    CTRL.Tem = 1.5 * CTRL.npp * CTRL.idq[1]*CTRL.KActive       # 电磁转矩计算
    if True:
        OB.iab_curr[0] = CTRL.iab[0]
        OB.iab_curr[1] = CTRL.iab[1]
        OB.uab_prev[0] = CTRL.cmd_uab[0] # Updating "OB.uab_prev[0]" here, which will be useful if voltage is measured, and is useless if assume command voltage is actual voltage.
        OB.uab_prev[1] = CTRL.cmd_uab[1] # Updating "OB.uab_prev[1]" here, which will be useful if voltage is measured, and is useless if assume command voltage is actual voltage.
        OB.uab_curr[0] = CTRL.cmd_uab[0]
        OB.uab_curr[1] = CTRL.cmd_uab[1]
        if False:
            RK4_OBSERVER_CJH_STYLE(DYNAMICS_2ndOdESO_LOF, OB.xspeed, OB, CTRL.CL_TS, CTRL)
        else:
            OB.output_error = np.sin(CTRL.theta_d - OB.xspeed[0]) # OE version 1
            RK4_OBSERVER_CJH_STYLE(DYNAMICS_4thOdESO_LOF, OB.xspeed, OB, CTRL.CL_TS, CTRL)
            # OB.output_error = np.sin(CTRL.theta_d - OB.xspeed[0]) # OE version 2
        while OB.xspeed[0]> np.pi: OB.xspeed[0] -= 2*np.pi
        while OB.xspeed[0]<-np.pi: OB.xspeed[0] += 2*np.pi
        OB.iab_prev[0] = OB.iab_curr[0]
        OB.iab_prev[1] = OB.iab_curr[1]
        # OB.uab_prev[0] = OB.uab_curr[0] # This is needed only if voltage is measured, e.g., by eCAP.
        # OB.uab_prev[1] = OB.uab_curr[1] # This is needed only if voltage is measured, e.g., by eCAP.

        """ Observer Outputs """
        OB.theta_d    = OB.xspeed[0]
        OB.omega_elec = OB.xspeed[1]
        if HUMAN.use_disturbance_feedforward_rejection == 0:
            OB.total_disrubance_feedforward = 0.0
        if HUMAN.use_disturbance_feedforward_rejection == 1:
            OB.total_disrubance_feedforward = OB.xspeed[2]
        elif HUMAN.use_disturbance_feedforward_rejection == 2:
            OB.total_disrubance_feedforward = OB.xspeed[2] + OB.ell2*OB.output_error

    """ Speed Measurement """
    if HUMAN.sensorless_speed:
        CTRL.omega_elec = OB.omega_elec
    else:
        CTRL.omega_elec = ACM.omega_elec

    """ (Optional) Do Park transformation again using the position estimate from the speed observer """

    """ Speed Controller """
    CTRL.cmd_udq = FOC(CTRL, reg_id, reg_iq, reg_speed, OB, HUMAN)
    # Inverse Park transformation
    CTRL.cmd_uab[0] = CTRL.cmd_udq[0] * CTRL.cosT + CTRL.cmd_udq[1] *-CTRL.sinT
    CTRL.cmd_uab[1] = CTRL.cmd_udq[0] * CTRL.sinT + CTRL.cmd_udq[1] * CTRL.cosT

    """ Inverter """
    ACM.uab[0] = CTRL.cmd_uab[0] # + current denpendent distorted voltage
    ACM.uab[1] = CTRL.cmd_uab[1] # + current denpendent distorted voltage

############################################# Wrapper level 1
""" MAIN for Real-time simulation """
@njit(nogil=True)
def ACMSimPyIncremental(
        t0, TIME,
        ACM=None,
        CTRL=None,
        reg_id=None,
        reg_iq=None,
        reg_speed=None,
        OB=None,
        HUMAN=None,
    ):
    MACHINE_TS = CTRL.CL_TS
    down_sampling_ceiling = int(CTRL.CL_TS / MACHINE_TS) #print('\tdown sample:', down_sampling_ceiling)

    # watch variabels
    machine_times  = np.arange(t0, t0+TIME, MACHINE_TS)
    control_times  = np.arange(t0, t0+TIME, CTRL.CL_TS)
    watch_data = np.zeros( (30, len(control_times)) )

    # Main loop
    # print('\tt0 =', t0)
    jj = 0; watch_index = 0
    for ii in range(len(machine_times)):

        t = machine_times[ii]

        """ Machine Simulation @ MACHINE_TS """
        # Park transformation
        ACM.udq[0] = ACM.uab[0] *  ACM.cosT + ACM.uab[1] * ACM.sinT
        ACM.udq[1] = ACM.uab[0] * -ACM.sinT + ACM.uab[1] * ACM.cosT
        # Numerical Integration (ode4) with 5 states
        RK4_MACHINE(t, ACM, hs=MACHINE_TS)

        """ Machine Simulation Output @ MACHINE_TS """
        # Generate output variables for easy access
        if ACM.x[3]> np.pi: ACM.x[3] -= 2*np.pi
        if ACM.x[3]<-np.pi: ACM.x[3] += 2*np.pi
        if ACM.x[4]> np.pi: ACM.x[4] -= 2*np.pi
        if ACM.x[4]<-np.pi: ACM.x[4] += 2*np.pi
        ACM.idq[0]      = ACM.x[0]
        ACM.idq[1]      = ACM.x[1]
        ACM.omega_elec  = ACM.x[2]
        ACM.theta_d     = ACM.x[3]
        ACM.theta_mech  = ACM.x[4]
        ACM.KActive     = ACM.x[5]
        # Inverse Park transformation
        ACM.cosT = np.cos(ACM.theta_d)
        ACM.sinT = np.sin(ACM.theta_d)
        ACM.iab[0] = ACM.x[0] * ACM.cosT + ACM.x[1] *-ACM.sinT
        ACM.iab[1] = ACM.x[0] * ACM.sinT + ACM.x[1] * ACM.cosT

        jj += 1
        if jj >= down_sampling_ceiling:
            jj = 0

            """ DSP @ CL_TS """
            DSP(ACM=ACM,
                CTRL=CTRL,
                reg_id=reg_id,
                reg_iq=reg_iq,
                reg_speed=reg_speed,
                OB=OB,
                HUMAN=HUMAN)

            """ Console @ CL_TS """
            if t < 1.0:
                CTRL.cmd_rpm = 50
            elif t < 1.5:
                ACM.TLoad = 2
            elif t < 2.0:
                CTRL.cmd_rpm = 200
            elif t < 3.0:
                CTRL.cmd_rpm = -200
            elif t < 4.0:
                CTRL.cmd_rpm = 0
            elif t < 4.5:
                CTRL.cmd_rpm = 2000
            elif t < 5:
                CTRL.cmd_idq[0] = 2
            elif t < 5.5:
                ACM.TLoad = 0.0
            elif t < 6: 
                HUMAN.CMD_SPEED_SINE_RPM = 500
            # else: # don't implement else to receive commands from IPython console

            if HUMAN.CMD_SPEED_SINE_RPM!=0:
                CTRL.cmd_rpm = HUMAN.CMD_SPEED_SINE_RPM * np.sin(2*np.pi*HUMAN.CMD_SPEED_SINE_HZ*t)

            """ Watch @ CL_TS """
            watch_data[ 0][watch_index] = ACM.x[0] # id
            watch_data[ 1][watch_index] = ACM.x[1] # iq
            watch_data[ 2][watch_index] = ACM.x[2] # omega_elec
            watch_data[ 3][watch_index] = ACM.x[3] # thete_d
            watch_data[ 4][watch_index] = ACM.x[4] # theta_mech
            watch_data[ 5][watch_index] = ACM.x[5] # KActive
            watch_data[ 6][watch_index] = CTRL.iab[0]
            watch_data[ 7][watch_index] = CTRL.iab[1]
            watch_data[ 8][watch_index] = CTRL.idq[0]
            watch_data[ 9][watch_index] = CTRL.idq[1]
            watch_data[10][watch_index] = CTRL.theta_d
            watch_data[11][watch_index] = CTRL.omega_elec / (2*np.pi*ACM.npp) * 60
            watch_data[12][watch_index] = CTRL.cmd_rpm
            watch_data[13][watch_index] = CTRL.cmd_idq[0]
            watch_data[14][watch_index] = CTRL.cmd_idq[1]
            watch_data[15][watch_index] = OB.xspeed[0] # theta_d
            watch_data[16][watch_index] = OB.xspeed[1] / (2*np.pi*ACM.npp) * 60 # omega_elec
            watch_data[17][watch_index] = OB.xspeed[2] # TL
            watch_data[18][watch_index] = OB.xspeed[3] # pT
            watch_data[19][watch_index] = CTRL.KActive
            watch_data[20][watch_index] = CTRL.KE
            watch_data[21][watch_index] = OB.xflux[0] # stator flux[0]
            watch_data[22][watch_index] = OB.xflux[1] # stator flux[1]
            watch_data[23][watch_index] = OB.xflux[2] # I term
            watch_data[24][watch_index] = OB.xflux[3] # I term
            watch_data[25][watch_index] = OB.active_flux[0] # active flux[0]
            watch_data[26][watch_index] = OB.active_flux[1] # active flux[1]
            watch_data[27][watch_index] = OB.theta_d # d-axis position
            watch_index += 1

    return control_times, watch_data
Watch_Mapping = [
    '[A]=ACM.x[0]', # id
    '[A]=ACM.x[1]', # iq
    '[rad/s]=ACM.x[2]', # omega_elec
    '[rad]=ACM.x[3]', # thete_d
    '[rad]=ACM.x[4]', # theta_mech
    '[Wb]=ACM.x[5]', # KActive
    '[A]=CTRL.iab[0]',
    '[A]=CTRL.iab[1]',
    '[A]=CTRL.idq[0]',
    '[A]=CTRL.idq[1]',
    '[rpm]=CTRL.theta_d',
    '[rpm]=CTRL.omega_elec',
    '[rpm]=CTRL.cmd_rpm',
    '[A]=CTRL.cmd_idq[0]',
    '[A]=CTRL.cmd_idq[1]',
    '[rad]=OB.xspeed[0]',  # theta_d
    '[rpm]=OB.xspeed[1]',  # omega_elec
    '[Nm]=OB.xspeed[2]',   # -TL
    '[Nm/s]=OB.xspeed[3]', # DL
    '[Wb]=CTRL.KActive',
    '[Wb]=CTRL.KE',
    '[Wb]=OB.xflux[0]', # stator flux[0]
    '[Wb]=OB.xflux[1]', # stator flux[1]
    '[V]=OB.xflux[2]', # I term
    '[V]=OB.xflux[3]', # I term
    '[Wb]=OB.active_flux[0]', # active flux[0]
    '[Wb]=OB.active_flux[1]', # active flux[1]
    '[rad]=OB.theta_d', # d-axis position
]

############################################# Wrapper level 2
def ACMSimPyWrapper(numba__scope_dict, *arg, **kwarg):

    # Do Numerical Integrations (that do not care about numba__scope_dict at all and return watch_data whatsoever)
    control_times, watch_data = ACMSimPyIncremental(*arg, **kwarg)

    # Post-processing
    numba__waveforms_dict = dict()
    for key, values in numba__scope_dict.items():
        # key = '$\alpha\beta$ current [A]'
        # values = ('CTRL.iab', 'CTRL.idq[1]'),
        waveforms = []
        for val in values:
            # val = 'CTRL.iab'
            for index, mapping in enumerate(Watch_Mapping):
                if val in mapping:
                    # CTRL.iab in '[A]=CTRL.iab[0]'
                    # CTRL.iab in '[A]=CTRL.iab[1]'
                    waveforms.append(watch_data[index])

                    # print('\t', key, val, 'in', mapping)
                    if len(val) == 1:
                        raise Exception('Invalid numba__scope_dict, make sure it is a dict of tuples of strings.')

        numba__waveforms_dict[key] = waveforms
    return control_times, numba__waveforms_dict



# %% Test incremental simulation
if __name__ == '__main__':

    # Simulation Basics
    CL_TS      = 1e-4
    TIME_SLICE = 1.0

    # Human
    HUMAN = The_Human()

    from collections import OrderedDict as OD
    numba__scope_dict = OD([
        (r'Speed [rpm]',                  ( 'CTRL.cmd_rpm', 'CTRL.omega_elec', 'OB.xspeed[1]'     ,) ),
        (r'Position [rad]',               ( 'ACM.x[3]', 'CTRL.theta_d', 'OB.xspeed[0]'            ,) ),
        (r'Position mech [rad]',          ( 'ACM.x[4]'                                       ,) ),
        (r'$q$-axis current [A]',         ( 'ACM.x[1]', 'CTRL.idq[1]', 'CTRL.cmd_idq[1]'     ,) ),
        (r'$d$-axis current [A]',         ( 'ACM.x[0]', 'CTRL.idq[0]', 'CTRL.cmd_idq[0]'     ,) ),
        (r'$\alpha\beta$ current [A]',    ( 'CTRL.iab'                        ,) ),
        (r'K_{\rm Active} [A]',           ( 'ACM.x[5]', 'CTRL.KActive'             ,) ),
        (r'Load torque [Nm]',             ( 'OB.xspeed[2]'                         ,) ),
    ])

    # init
    ACM       = The_AC_Machines(HUMAN)
    CTRL      = The_Motor_Controller(CL_TS, 5*CL_TS, HUMAN)
    reg_id    = The_PI_Regulator(6.39955, 6.39955*237.845*CTRL.CL_TS, 600)
    reg_iq    = The_PI_Regulator(6.39955, 6.39955*237.845*CTRL.CL_TS, 600)
    reg_speed = The_PI_Regulator(0.0380362, 0.0380362*30.5565*CTRL.VL_TS, 1*1.414*ACM.IN)
    OB        = The_Observer(HUMAN)

    # Global arrays
    global_ACM_speed, global__OB_speed = None, None
    global___KActive = None
    global____ACM_id, global___CTRL_id = None, None
    global________x5 = None
    global_CTRL_theta_d, global_OB_theta_d = None, None
    global_TL = None

    for ii in range(0, 10):
        """perform animation step"""
        control_times, numba__waveforms_dict = \
            ACMSimPyWrapper(numba__scope_dict,
                        t0=ii*TIME_SLICE, TIME=TIME_SLICE, 
                        ACM=ACM,
                        CTRL=CTRL,
                        reg_id=reg_id,
                        reg_iq=reg_iq,
                        reg_speed=reg_speed,
                        OB=OB,
                        HUMAN=HUMAN)
        cmd_rpm, ACM_speed, OB_speed = numba__waveforms_dict[r'Speed [rpm]']
        x5, KActive                  = numba__waveforms_dict[r'K_{\rm Active} [A]']
        ACM_id, CTRL_id, cmd_id      = numba__waveforms_dict[r'$d$-axis current [A]']
        _, CTRL_theta_d, OB_theta_d  = numba__waveforms_dict[r'Position [rad]']
        TL,                          = numba__waveforms_dict[r'Load torque [Nm]']

        def save_to_global(_global, _local):
            return _local if _global is None else np.append(_global, _local)    
        global_ACM_speed = save_to_global(global_ACM_speed, ACM_speed)
        global__OB_speed = save_to_global(global__OB_speed,  OB_speed)
        global________x5 = save_to_global(global________x5,        x5)
        global___KActive = save_to_global(global___KActive,   KActive)
        global____ACM_id = save_to_global(global____ACM_id,    ACM_id)
        global___CTRL_id = save_to_global(global___CTRL_id,   CTRL_id)
        global_CTRL_theta_d = save_to_global(global_CTRL_theta_d, CTRL_theta_d)
        global_OB_theta_d = save_to_global(global_OB_theta_d, OB_theta_d)
        global_TL        = save_to_global(global_TL, TL)

        print(ACM.KActive, 'Wb')

        # for k,v in numba__waveforms_dict.items():
        #     print(k, np.shape(v))

        # print(len(global_speed), end='|')
        # print(max(global_speed), end='|')
        # print(len(speed), end='|')
        # print(max(speed))
        # print()
        # break

    plt.figure(figsize=(15,4))
    plt.plot(global_ACM_speed); plt.plot(global__OB_speed)

    plt.figure(figsize=(15,4)); plt.ylim([-1e-1, 1e-1])
    plt.plot( (global_ACM_speed - global__OB_speed) ) # [2000:5000]

    plt.figure(figsize=(15,4))
    plt.plot( global________x5 ); plt.plot( global___KActive )

    plt.figure(figsize=(15,4))
    plt.plot( global____ACM_id ); plt.plot( global___CTRL_id )

    plt.figure(figsize=(15,4))
    plt.plot( global_CTRL_theta_d ); plt.plot( global_OB_theta_d )

    plt.figure(figsize=(15,4))
    plt.plot( np.sin(global_CTRL_theta_d - global_OB_theta_d) )

    plt.figure(figsize=(15,4))
    plt.plot( global_TL )

    print(OB.ell1, OB.ell2, OB.ell3, OB.ell4)



# %%



# %%
import sys; 
sys.path.insert(1, r'D:\DrH\Codes\emachineryTestPYPI\emachinery\acmsimcv5\CJHAcademicPlot')
from __cjhAcademicPlotSettings import CJHStylePlot, pd, np, plt, os
sys.path.insert(1, r'D:\DrH\Codes\visualize')
from DFT_CJH import myDFT_single_channel, DFT_CJH

# %%
''' Current Function
'''
I_AMPL = 1.0
PERIOD = 0.02 # s
CL_TS = 1e-4
time = np.arange(0, PERIOD, CL_TS)
phase_angle = np.arange(-np.pi, np.pi, 2*np.pi / len(time))

def sigmoid(x, a1=0, a2=1, a3=10):
    return a1 * x + a2* (2 / (1.0 + np.exp(-a3 * x)) - 1)

""" Compensation Voltage Function By Current Value """

Ix = np.arange(0,2,0.01)
plt.figure(figsize=(20,6))
plt.plot(Ix, sigmoid(Ix, a2=4, a3=5))
plt.plot(Ix, sigmoid(Ix, a2=2, a3=10))
plt.plot(Ix, sigmoid(Ix, a2=1, a3=20))


#%% In time domain

def main():
    dict_label_dataframe = dict()

    I_AMPL_list = [1, 0.2]
    # I_AMPL_list = [1, ]
    # I_AMPL_list = [5, ]
    for key, I_AMPL in zip([str(el) for el in I_AMPL_list], I_AMPL_list):
        phase_current = np.sin(phase_angle) * I_AMPL
        ui_curve = [sigmoid(current_value) for current_value in phase_current]
        dict_label_dataframe.update(
            {
                key: [
                    pd.DataFrame({
                        'Time': time,
                        'Signal': phase_current
                    }), 
                    pd.DataFrame({
                        'Time': time,
                        'Signal': ui_curve
                    }), 
                ] # 如果不是一个list，则CSP会报错：TypeError: string indices must be integers
            }
        )

    def adjust_figure(axes_v):
        axes_v[0].set_ylim([-1.1,1.1])
        axes_v[1].set_ylim([-1.1,1.1])
    def adjust_ylabel(axes_v):
        axes_v[0].set_ylabel('[A] or [V]')
        axes_v[1].set_ylabel(None)
        # for ax in axes_v:
        #     ax.set_ylabel('[A] or [V]')

        axes_v[0].set_xlabel('Time, $t$ [s]\n(a)')
        axes_v[1].set_xlabel('Time, $t$ [s]\n(b)')

    csp = CJHStylePlot( ncols=len(I_AMPL_list), figsize=(5, 1.5)) #figsize=(4, 1.5) )
    csp.aca_plot(dict_label_dataframe, x='Time', y='Signal', 
                    adjust_figure=adjust_figure,
                    adjust_ylabel=adjust_ylabel)
    csp.aca_save(f'{os.path.dirname(__file__)}/compensation_voltage_shape', bool_sumatrapdf=True)

    return ui_curve, csp

ui_curve, csp = main()


# %%
""" 检查 sigmoid in time domain 随着 Ix 和 omega 的变化而变化的规律 """

sigma = lambda t, a3, Ix, omega: 2 / (1 + np.exp(-a3*Ix*np.sin(omega*t))) - 1

Ix = 2
omega = 50*2*np.pi

t = np.arange(0, 0.02, CL_TS)
ix = Ix*np.sin(omega*t)
Dx = sigma(t, a3=10, Ix=Ix, omega=omega)
plt.plot(t, ix)
plt.plot(t, Dx)

# %%

def basic_DFT_analysis(raw_data, Frequency):
    REPEAT = 10

    time = np.arange(0, REPEAT/Frequency, CL_TS)

    signal = np.array(REPEAT*list(raw_data))
    ax1 = ax_matrix[0]
    ax2 = ax_matrix[1]
    ax1.set_xlabel('Frequency [Hz]')
    ax1.set_ylabel('Amplitude [dB]')
    ax2.set_xlabel('Frequency [Hz]')
    ax2.set_ylabel('Phase [deg]')

    DFT_CJH(ax1, time, signal, bool_use_dB=False, ax2=ax2)

    ax1.set_xlim([0,200])
    # ax.set_ylim([-50,5])
    ax1.grid(1)
    ax2.grid(1)

    print(len(time), len(signal))

sigma = lambda t, a3, Ix, omega: 2 / (1 + np.exp(-a3*Ix*np.sin(omega*t))) - 1
Frequency = 10
t = np.arange(0, 1/Frequency, CL_TS)
fig, ax_matrix = plt.subplots(ncols=1, nrows=2, sharex=True, figsize=(20,6))
for Ix in [0.1, 0.5, 10]:
    Dx = sigma(t, a3=10, Ix=Ix, omega=Frequency*2*np.pi)
    basic_DFT_analysis(Dx, Frequency)


# %%

def parametric_analysis_of_harmonic_content(raw_data, Frequency):
    REPEAT = 1
    time = np.arange(0, REPEAT/Frequency, CL_TS)
    signal = np.array(REPEAT*list(raw_data))
    dft_dict = myDFT_single_channel(time, signal)

    if REPEAT == 1: # 如果REPEAT变了分辨率就变了，间隔也就变了哦。
        harmonics[ 1].append(dft_dict['ampl'][ 1]); #print(' 10 Hz', dft_dict['ampl'][ 1])
        harmonics[ 5].append(dft_dict['ampl'][ 5]); #print(' 50 Hz', dft_dict['ampl'][ 5])
        harmonics[ 7].append(dft_dict['ampl'][ 7]); #print(' 70 Hz', dft_dict['ampl'][ 7])
        harmonics[11].append(dft_dict['ampl'][11]); #print('110 Hz', dft_dict['ampl'][11])
        harmonics[13].append(dft_dict['ampl'][13]); #print('130 Hz', dft_dict['ampl'][13])

# the problem
sigma = lambda t, a3, Ix, omega: 2 / (1 + np.exp(-a3*Ix*np.sin(omega*t))) - 1
Frequency = 10
CL_TS = 1e-4
t = np.arange(0, 1/Frequency, CL_TS)

# init
def init_harmonics_dict():
    harmonics = dict()
    for h in [1,5,7,11,13]:
        harmonics[h] = []
    return harmonics

a2_list = [1, 2, 4]
for a2 in a2_list:

    # iterate through different a3 values
    a3_list = [5,10,20,40]
    for a3 in a3_list:

        harmonics = init_harmonics_dict()

        # iterate through current peak value
        # Ix_list = [0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4] #, 12.8, 25.6, 51.2, 102.4]
        Ix_list = np.arange(0.1, 6.31, 0.05) #, 12.8, 25.6, 51.2, 102.4]
        Ix_list = np.arange(0.1, 3, 0.05)
        for Ix in Ix_list:
            Dx = a2*sigma(t, a3=a3, Ix=Ix, omega=Frequency*2*np.pi)
            parametric_analysis_of_harmonic_content(Dx, Frequency)

        plt.figure(1, figsize=(8,6), facecolor='w')
        plt.plot(Ix_list, harmonics[1], '.--', label=f'$a_2={a2},~a_3={a3}$') # marker='$1$', 
        plt.legend()
        plt.title('Amplitude of the Fundamental Component in $\hat D_x(t)$')
        plt.grid(1)

        plt.figure(5, figsize=(20,6), facecolor='w')
        plt.plot(Ix_list, harmonics[5], '.--', label=f'$a_2={a2},~a_3={a3}$') # marker='$5$', 
        plt.legend()
        plt.title('5th order harmonic')
        plt.grid(1)

        plt.figure(11, figsize=(20,6), facecolor='w')
        plt.plot(Ix_list, harmonics[11], '.--', label=f'$a_2={a2},~a_3={a3}$') # marker='$11$', 
        plt.legend()
        plt.title('11th order harmonic')
        plt.grid(1)

        # plt.figure(7, figsize=(20,6), facecolor='w')
        # plt.plot(Ix_list, harmonics[7], '.-', label=f'$a_2={a2},~a_3={a3}$')
        # plt.legend()
        # plt.title('7th order harmonic')
        # plt.grid(1)

        # plt.figure(13, figsize=(20,6), facecolor='w')
        # plt.plot(Ix_list, harmonics[13], '.-', label=f'$a_2={a2},~a_3={a3}$')
        # plt.legend()
        # plt.title('13th order harmonic')
        # plt.grid(1)

        print(Ix_list[int(sum(np.array(harmonics[1])<1.209546*a2))-1] * a3)

    fig = plt.figure(1)
    plt.plot(Ix_list, np.zeros_like(Ix_list) + harmonics[1][-1] * 0.95, 'k--')
    plt.ylabel('Voltage [V]')
    plt.xlabel('Current Peak Value, $I_x$ [A]')
    # print(harmonics[1][-1] * 0.95)

csp.aca_save(f'{os.path.dirname(__file__)}/harmonic_contents', bool_sumatrapdf=True, fig=fig)


# %%

""" Analysis with Park&Sul Style 
    Voltage components as a function of a3Ix.
"""

fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(9,3), facecolor='w', sharey=True)

a3Ix_list = np.arange(0, 80, 0.5)

for a2, alpha in zip([3,2,1], [1, 0.7, 0.5]):

    harmonics = init_harmonics_dict()

    for a3Ix in a3Ix_list:

        # iterate through current peak value
        Dx = a2*sigma(t, a3=a3Ix, Ix=1, omega=Frequency*2*np.pi)
        parametric_analysis_of_harmonic_content(Dx, Frequency)

    print(a3Ix_list[int(sum(np.array(harmonics[1])<1.209546*a2))-1] * 1) # 1.209546 is 95% of the max fundamental

    ax = axes[0]
    ax.plot(a3Ix_list, harmonics[1], '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.title.set_text('Fundamental')
    ax.grid(1)
    ax.set_xlabel('$a_3I_x$ [1]')

    ax = axes[1]
    ax.plot(a3Ix_list, np.array(harmonics[5]), '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.title.set_text('5th order harmonic')
    ax.grid(1)
    ax.set_xlabel('$a_3I_x$ [1]')

    ax = axes[2]
    ax.plot(a3Ix_list, np.array(harmonics[11]), '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.title.set_text('11th order harmonic')
    ax.grid(1)
    ax.set_xlabel('$a_3I_x$ [1]')

    # plt.figure(7, figsize=(20,6), facecolor='w')
    # plt.plot(a3Ix_list, harmonics[7], '.-', label=f'$a_2={a2}$'
    # plt.legend()
    # plt.title('7th order harmonic')
    # plt.grid(1)
    # plt.xlabel('$a_3I_x$')

    # plt.figure(13, figsize=(20,6), facecolor='w')
    # plt.plot(a3Ix_list, harmonics[13], '.-', label=f'$a_2={a2}$'
    # plt.legend()
    # plt.title('13th order harmonic')
    # plt.grid(1)
    # plt.xlabel('$a_3I_x$')

    axes[0].plot(a3Ix_list, np.zeros_like(a3Ix_list) + harmonics[1][-1] * 0.95, 'k--')

axes[0].set_ylabel('Voltage [V]')
plt.plot([], [], ' ', label="From top to bottom")
plt.legend()

csp.aca_save(f'{os.path.dirname(__file__)}/harmonic_contents_by_a3Ix', bool_sumatrapdf=True, fig=fig)

#%%

""" 重制，谐波和基波的比例既然不变，这个信息就要利用起来。
"""
import matplotlib.ticker as mtick

#fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(8,3), facecolor='w', sharey=False)
fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10,3), facecolor='w', sharey=False)
fig.subplots_adjust(wspace=0.5)

a3Ix_list = np.arange(0.5, 80, 0.5)

for a2, alpha in zip([3,2,1], [1, 0.7, 0.5]):

    harmonics = init_harmonics_dict()

    for a3Ix in a3Ix_list:

        # iterate through current peak value
        Dx = a2*sigma(t, a3=a3Ix, Ix=1, omega=Frequency*2*np.pi)
        parametric_analysis_of_harmonic_content(Dx, Frequency)

    print(a3Ix_list[int(sum(np.array(harmonics[1])<1.209546*a2))-1] * 1) # 1.209546 is 95% of the max fundamental

    ax = axes[0]
    ax.plot(a3Ix_list, harmonics[1], '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.title.set_text('Fundamental')
    ax.grid(1)
    ax.set_xlabel('$a_3I_x^*$ [1]\n(a)')

    ax = axes[1]
    ax.plot(a3Ix_list, np.array(harmonics[5]) / np.array(harmonics[1]), '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.plot(a3Ix_list, np.array(harmonics[11]) / np.array(harmonics[1]), '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.plot(a3Ix_list, np.array(harmonics[7]) / np.array(harmonics[1]), '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.plot(a3Ix_list, np.array(harmonics[13]) / np.array(harmonics[1]), '-', label=f'$a_2={a2}$', alpha=alpha)
    ax.title.set_text('Harmonics')
    ax.grid(1)
    ax.set_xlabel('$a_3I_x^*$ [1]\n(b)')

    axes[0].plot(a3Ix_list, np.zeros_like(a3Ix_list) + harmonics[1][-1] * 0.95, 'k--')

axes[0].set_ylabel('Voltage [V]')
axes[1].set_ylabel('Harmonic to\nFundamental Ratio [%]')
plt.plot([], [], ' ', label="From top to bottom")
# plt.legend()
axes[1].yaxis.set_major_formatter(mtick.PercentFormatter(xmax=1.0, decimals=0))

csp.aca_save(f'{os.path.dirname(__file__)}/harmonic_contents_by_a3Ix_ratio', bool_sumatrapdf=True, fig=fig)

# %%



import time, serial, sys

# configure the serial connections (the parameters differs on the device you are connecting to)
console = serial.Serial(
    port='COM5',
    baudrate=9600,
    parity=serial.PARITY_NONE,
    stopbits=serial.STOPBITS_ONE,
    bytesize=serial.EIGHTBITS,
)
console.isOpen() # console.open()
console.write(b'hello')

user_input=1
while 1 :
    # get keyboard user_input
    # user_input = sys.stdin.read()
    user_input = input(">> ")
    print("USER", user_input)
    if user_input == 'exit':
        console.close()
    else:
        # send the character to the device
        # (note that I happend a \r\n carriage return and line feed to the characters - this is requested by my device)
        console.write(user_input.encode()) # default encoding to utf-8
        out = []

        # let's wait one second before reading output (let's give device time to answer)
        time.sleep(1)
        while console.inWaiting() > 0:
            out.append( console.read(1) )

        if out != []:
            print("DSP:" , out)




# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT QT CORE
# ///////////////////////////////////////////////////////////////
from qt_core import *

# STYLE
# ///////////////////////////////////////////////////////////////
style = '''
QPushButton {{
	border: none;
    padding-left: 10px;
    padding-right: 5px;
    color: {_color};
	border-radius: {_radius};	
	background-color: {_bg_color};
}}
QPushButton:hover {{
	background-color: {_bg_color_hover};
}}
QPushButton:pressed {{	
	background-color: {_bg_color_pressed};
}}
'''

# PY PUSH BUTTON
# ///////////////////////////////////////////////////////////////
class PyPushButton(QPushButton):
    def __init__(
        self, 
        text,
        radius,
        color,
        bg_color,
        bg_color_hover,
        bg_color_pressed,
        parent = None,
    ):
        super().__init__()

        # SET PARAMETRES
        self.setText(text)
        if parent != None:
            self.setParent(parent)
        self.setCursor(Qt.PointingHandCursor)

        # SET STYLESHEET
        custom_style = style.format(
            _color = color,
            _radius = radius,
            _bg_color = bg_color,
            _bg_color_hover = bg_color_hover,
            _bg_color_pressed = bg_color_pressed
        )
        self.setStyleSheet(custom_style)

        


# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT QT CORE
# ///////////////////////////////////////////////////////////////
from qt_core import *

# IMPORT STYLE
# ///////////////////////////////////////////////////////////////
from . style import *

# PY PUSH BUTTON
# ///////////////////////////////////////////////////////////////
class PyTableWidget(QTableWidget):
    def __init__(
        self, 
        radius = 8,
        color = "#FFF",
        bg_color = "#444",
        selection_color = "#FFF",
        header_horizontal_color = "#333",
        header_vertical_color = "#444",
        bottom_line_color = "#555",
        grid_line_color = "#555",
        scroll_bar_bg_color = "#FFF",
        scroll_bar_btn_color = "#3333",
        context_color = "#00ABE8"
    ):
        super().__init__()

        # PARAMETERS

        # SET STYLESHEET
        self.set_stylesheet(
            radius,
            color,
            bg_color,
            header_horizontal_color,
            header_vertical_color,
            selection_color,
            bottom_line_color,
            grid_line_color,
            scroll_bar_bg_color,
            scroll_bar_btn_color,
            context_color
        )

    # SET STYLESHEET
    def set_stylesheet(
        self,
        radius,
        color,
        bg_color,
        header_horizontal_color,
        header_vertical_color,
        selection_color,
        bottom_line_color,
        grid_line_color,
        scroll_bar_bg_color,
        scroll_bar_btn_color,
        context_color
    ):
        # APPLY STYLESHEET
        style_format = style.format(
            _radius = radius,          
            _color = color,
            _bg_color = bg_color,
            _header_horizontal_color = header_horizontal_color,
            _header_vertical_color = header_vertical_color,
            _selection_color = selection_color,
            _bottom_line_color = bottom_line_color,
            _grid_line_color = grid_line_color,
            _scroll_bar_bg_color = scroll_bar_bg_color,
            _scroll_bar_btn_color = scroll_bar_btn_color,
            _context_color = context_color
        )
        self.setStyleSheet(style_format)


# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT QT CORE
# ///////////////////////////////////////////////////////////////
from qt_core import *


# IMPORT PACKAGES AND MODULES
# ///////////////////////////////////////////////////////////////
from gui.uis.windows.main_window.functions_main_window import MainFunctions, Functions #from . ui_main import Functions
import os, sys

# IMPORT SETTINGS
# ///////////////////////////////////////////////////////////////
from gui.core.json_settings import Settings

# IMPORT PY ONE DARK WINDOWS
# ///////////////////////////////////////////////////////////////
# MAIN WINDOW
from gui.uis.windows.main_window import *

# IMPORT PY ONE DARK WIDGETS
# ///////////////////////////////////////////////////////////////
from gui.widgets import *


# IMPORT EMY Functions
# ///////////////////////////////////////////////////////////////
from emy_functions import EmyFunctions


# ADJUST QT FONT DPI FOR HIGHT SCALE AN 4K MONITOR
# ///////////////////////////////////////////////////////////////
os.environ["QT_FONT_DPI"] = "96"
# IF IS 4K MONITOR ENABLE 'os.environ["QT_SCALE_FACTOR"] = "2"'

# MAIN WINDOW
# ///////////////////////////////////////////////////////////////
class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        # SETUP MAIN WINDOW
        # Load widgets from "gui\uis\main_window\ui_main.py"
        # ///////////////////////////////////////////////////////////////
        self.ui = UI_MainWindow()
        self.ui.setup_ui(self)

        # LOAD SETTINGS
        # ///////////////////////////////////////////////////////////////
        settings = Settings()
        self.settings = settings.items

        # SETUP MAIN WINDOW
        # ///////////////////////////////////////////////////////////////
        self.hide_grips = True # Show/Hide resize grips
        SetupMainWindow.setup_gui(self)



        # SHOW MAIN WINDOW
        # ///////////////////////////////////////////////////////////////
        self.show()

    # LEFT MENU BTN IS CLICKED
    # Run function when btn is clicked
    # Check funtion by object name / btn_id
    # ///////////////////////////////////////////////////////////////
    def btn_clicked(self):
        # GET BT CLICKED
        btn = SetupMainWindow.setup_btns(self)

        # Remove Selection If Clicked By "btn_close_left_column"
        if btn.objectName() != "btn_settings":
            self.ui.left_menu.deselect_all_tab()

        # Get Title Bar Btn And Reset Active         
        top_settings = MainFunctions.get_title_bar_btn(self, "btn_top_settings")
        top_settings.set_active(False)

        # LEFT MENU
        # ///////////////////////////////////////////////////////////////
        
        # HOME BTN
        if btn.objectName() == "btn_home":
            # Select Menu
            self.ui.left_menu.select_only_one(btn.objectName())

            # Load Page 1
            MainFunctions.set_page(self, self.ui.load_pages.page_1)

        # WIDGETS BTN
        if btn.objectName() == "btn_widgets":
            # Select Menu
            self.ui.left_menu.select_only_one(btn.objectName())

            # Load Page 2
            MainFunctions.set_page(self, self.ui.load_pages.page_2)

        # LOAD USER PAGE
        if btn.objectName() == "btn_add_user":
            # Select Menu
            self.ui.left_menu.select_only_one(btn.objectName())

            # Load Page 3 
            MainFunctions.set_page(self, self.ui.load_pages.page_3)

        # LOAD USER PAGE
        if btn.objectName() == "btn_scope":
            # Select Menu
            self.ui.left_menu.select_only_one(btn.objectName())

            # Load Page 4 Scope
            MainFunctions.set_page(self, self.ui.load_pages.page_4_scope)




        # BOTTOM INFORMATION
        if btn.objectName() == "btn_info":
            # CHECK IF LEFT COLUMN IS VISIBLE
            if not MainFunctions.left_column_is_visible(self):
                self.ui.left_menu.select_only_one_tab(btn.objectName())

                # Show / Hide
                MainFunctions.toggle_left_column(self)
                self.ui.left_menu.select_only_one_tab(btn.objectName())
            else:
                if btn.objectName() == "btn_close_left_column":
                    self.ui.left_menu.deselect_all_tab()
                    # Show / Hide
                    MainFunctions.toggle_left_column(self)
                
                self.ui.left_menu.select_only_one_tab(btn.objectName())

            # Change Left Column Menu
            if btn.objectName() != "btn_close_left_column":
                MainFunctions.set_left_column_menu(
                    self, 
                    menu = self.ui.left_column.menus.menu_2,
                    title = "Info tab",
                    icon_path = Functions.set_svg_icon("icon_info.svg")
                )



        # SETTINGS LEFT
        if btn.objectName() == "btn_settings" or btn.objectName() == "btn_close_left_column":
            # CHECK IF LEFT COLUMN IS VISIBLE
            if not MainFunctions.left_column_is_visible(self):
                # Show / Hide
                MainFunctions.toggle_left_column(self)
                self.ui.left_menu.select_only_one_tab(btn.objectName())
            else:
                if btn.objectName() == "btn_close_left_column":
                    self.ui.left_menu.deselect_all_tab()
                    # Show / Hide
                    MainFunctions.toggle_left_column(self)
                self.ui.left_menu.select_only_one_tab(btn.objectName())

            # Change Left Column Menu
            if btn.objectName() != "btn_close_left_column":
                MainFunctions.set_left_column_menu(
                    self, 
                    menu = self.ui.left_column.menus.menu_1,
                    title = "Settings Left Column",
                    icon_path = Functions.set_svg_icon("icon_settings.svg")
                )
        


        # TITLE BAR MENU
        # ///////////////////////////////////////////////////////////////
        
        # SETTINGS TITLE BAR
        if btn.objectName() == "btn_top_settings":
            # Toogle Active
            if not MainFunctions.right_column_is_visible(self):
                btn.set_active(True)

                # Show / Hide
                MainFunctions.toggle_right_column(self)
            else:
                btn.set_active(False)

                # Show / Hide
                MainFunctions.toggle_right_column(self)

            # Get Left Menu Btn            
            top_settings = MainFunctions.get_left_menu_btn(self, "btn_settings")
            top_settings.set_active_tab(False)            

        if btn.objectName() == "btn_run_realtime_simulation":

            # switch to scope page
            MainFunctions.set_page(self, self.ui.load_pages.page_4_scope)
            EmyFunctions.runPyBasedSimulation(self, bool_realtime=True)

        # DEBUG
        print(f"Button {btn.objectName()}, clicked!")

    # LEFT MENU BTN IS RELEASED
    # Run function when btn is released
    # Check funtion by object name / btn_id
    # ///////////////////////////////////////////////////////////////
    def btn_released(self):
        # GET BT CLICKED
        btn = SetupMainWindow.setup_btns(self)

        # DEBUG
        print(f"Button {btn.objectName()}, released!")

    # RESIZE EVENT
    # ///////////////////////////////////////////////////////////////
    def resizeEvent(self, event):
        SetupMainWindow.resize_grips(self)

    # MOUSE CLICK EVENTS
    # ///////////////////////////////////////////////////////////////
    def mousePressEvent(self, event):
        # SET DRAG POS WINDOW
        self.dragPos = event.globalPos()




    # 把变量推到 qtconsole 中去
    def console_push_variable(self, d):
        self.ui.right_column.ConsoleWidget_1.push_vars(d)

        # 显示哪些变量已经被推到 qtconsole 中去了：
        # if self.ui.label_pushedVariables.text() == 'None':
        #     self.ui.label_pushedVariables.setText(', '.join(d.keys()))
        # else:
        #     # Duplicated variable names would exist
        #     # self.ui.label_pushedVariables.setText(self.ui.label_pushedVariables.text()+', '+', '.join(d.keys()))

        #     # Avoid duplicated variable names
        #     for key in d.keys():
        #         if key not in self.ui.label_pushedVariables.text():
        #             self.ui.label_pushedVariables.setText(self.ui.label_pushedVariables.text()+f', {key}')



# SETTINGS WHEN TO START
# Set the initial class and also additional parameters of the "QApplication" class
# ///////////////////////////////////////////////////////////////
if __name__ == "__main__":
    # APPLICATION
    # ///////////////////////////////////////////////////////////////
    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon("icon.ico"))
    window = MainWindow()

    # EXEC APP
    # ///////////////////////////////////////////////////////////////
    sys.exit(app.exec_())



# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT PACKAGES AND MODULES
# ///////////////////////////////////////////////////////////////
import os

# IMPORT QT CORE
# ///////////////////////////////////////////////////////////////
from qt_core import *

# IMPORT FUNCTIONS
# ///////////////////////////////////////////////////////////////
from gui.core.functions import *

# CUSTOM LEFT MENU
# ///////////////////////////////////////////////////////////////
class PyLeftMenuButton(QPushButton):
    def __init__(
        self,
        app_parent,
        text,
        btn_id = None,
        tooltip_text = "",
        margin = 4,
        dark_one = "#1b1e23",
        dark_three = "#21252d",
        dark_four = "#272c36",
        bg_one = "#2c313c",
        icon_color = "#c3ccdf",
        icon_color_hover = "#dce1ec",
        icon_color_pressed = "#edf0f5",
        icon_color_active = "#f5f6f9",
        context_color = "#568af2",
        text_foreground = "#8a95aa",
        text_active = "#dce1ec",
        icon_path = "icon_add_user.svg",
        icon_active_menu = "active_menu.svg",
        is_active = False,
        is_active_tab = False,
        is_toggle_active = False
    ):
        super().__init__()
        self.setText(text)
        self.setCursor(Qt.PointingHandCursor)
        self.setMaximumHeight(50)
        self.setMinimumHeight(50)
        self.setObjectName(btn_id)

        # APP PATH
        self._icon_path = Functions.set_svg_icon(icon_path)
        self._icon_active_menu = Functions.set_svg_icon(icon_active_menu)

        # PROPERTIES
        self._margin = margin
        self._dark_one = dark_one
        self._dark_three = dark_three
        self._dark_four = dark_four
        self._bg_one = bg_one
        self._context_color = context_color
        self._icon_color = icon_color
        self._icon_color_hover = icon_color_hover
        self._icon_color_pressed = icon_color_pressed
        self._icon_color_active = icon_color_active
        self._set_icon_color = self._icon_color # Set icon color
        self._set_bg_color = self._dark_one # Set BG color
        self._set_text_foreground = text_foreground
        self._set_text_active = text_active
        self._parent = app_parent
        self._is_active = is_active
        self._is_active_tab = is_active_tab
        self._is_toggle_active = is_toggle_active

        # TOOLTIP
        self._tooltip_text = tooltip_text
        self.tooltip = _ToolTip(
            app_parent,
            tooltip_text,
            dark_one,
            context_color,
            text_foreground
        )
        self.tooltip.hide()

    # PAINT EVENT
    # ///////////////////////////////////////////////////////////////
    def paintEvent(self, event):
        # PAINTER
        p = QPainter()
        p.begin(self)
        p.setRenderHint(QPainter.Antialiasing)
        p.setPen(Qt.NoPen)
        p.setFont(self.font())

        # RECTANGLES
        rect = QRect(4, 5, self.width(), self.height() - 10)
        rect_inside = QRect(4, 5, self.width() - 8, self.height() - 10)
        rect_icon = QRect(0, 0, 50, self.height())
        rect_blue = QRect(4, 5, 20, self.height() - 10)
        rect_inside_active = QRect(7, 5, self.width(), self.height() - 10)
        rect_text = QRect(45, 0, self.width() - 50, self.height())

        if self._is_active:
            # DRAW BG BLUE
            p.setBrush(QColor(self._context_color))
            p.drawRoundedRect(rect_blue, 8, 8)

            # BG INSIDE
            p.setBrush(QColor(self._bg_one))
            p.drawRoundedRect(rect_inside_active, 8, 8)

            # DRAW ACTIVE
            icon_path = self._icon_active_menu
            app_path = os.path.abspath(os.getcwd())
            icon_path = os.path.normpath(os.path.join(app_path, icon_path))
            self._set_icon_color = self._icon_color_active
            self.icon_active(p, icon_path, self.width())

            # DRAW TEXT
            p.setPen(QColor(self._set_text_active))
            p.drawText(rect_text, Qt.AlignVCenter, self.text())

            # DRAW ICONS
            self.icon_paint(p, self._icon_path, rect_icon, self._set_icon_color) 

        elif self._is_active_tab:
            # DRAW BG BLUE
            p.setBrush(QColor(self._dark_four))
            p.drawRoundedRect(rect_blue, 8, 8)

            # BG INSIDE
            p.setBrush(QColor(self._bg_one))
            p.drawRoundedRect(rect_inside_active, 8, 8)

            # DRAW ACTIVE
            icon_path = self._icon_active_menu
            app_path = os.path.abspath(os.getcwd())
            icon_path = os.path.normpath(os.path.join(app_path, icon_path))
            self._set_icon_color = self._icon_color_active
            self.icon_active(p, icon_path, self.width())

            # DRAW TEXT
            p.setPen(QColor(self._set_text_active))
            p.drawText(rect_text, Qt.AlignVCenter, self.text())

            # DRAW ICONS
            self.icon_paint(p, self._icon_path, rect_icon, self._set_icon_color)

        # NORMAL BG
        else:
            if self._is_toggle_active:
                # BG INSIDE
                p.setBrush(QColor(self._dark_three))
                p.drawRoundedRect(rect_inside, 8, 8)

                # DRAW TEXT
                p.setPen(QColor(self._set_text_foreground))
                p.drawText(rect_text, Qt.AlignVCenter, self.text())

                # DRAW ICONS
                if self._is_toggle_active:
                    self.icon_paint(p, self._icon_path, rect_icon, self._context_color)
                else:
                    self.icon_paint(p, self._icon_path, rect_icon, self._set_icon_color)
            else:
                # BG INSIDE
                p.setBrush(QColor(self._set_bg_color))
                p.drawRoundedRect(rect_inside, 8, 8)

                # DRAW TEXT
                p.setPen(QColor(self._set_text_foreground))
                p.drawText(rect_text, Qt.AlignVCenter, self.text())

                # DRAW ICONS
                self.icon_paint(p, self._icon_path, rect_icon, self._set_icon_color)        

        p.end()

    # SET ACTIVE MENU
    # ///////////////////////////////////////////////////////////////
    def set_active(self, is_active):
        self._is_active = is_active
        if not is_active:
            self._set_icon_color = self._icon_color
            self._set_bg_color = self._dark_one

        self.repaint()

    # SET ACTIVE TAB MENU
    # ///////////////////////////////////////////////////////////////
    def set_active_tab(self, is_active):
        self._is_active_tab = is_active
        if not is_active:
            self._set_icon_color = self._icon_color
            self._set_bg_color = self._dark_one
            
        self.repaint()

    # RETURN IF IS ACTIVE MENU
    # ///////////////////////////////////////////////////////////////
    def is_active(self):
        return self._is_active

    # RETURN IF IS ACTIVE TAB MENU
    # ///////////////////////////////////////////////////////////////
    def is_active_tab(self):
        return self._is_active_tab
    
    # SET ACTIVE TOGGLE
    # ///////////////////////////////////////////////////////////////
    def set_active_toggle(self, is_active):
        self._is_toggle_active = is_active

    # SET ICON
    # ///////////////////////////////////////////////////////////////
    def set_icon(self, icon_path):
        self._icon_path = icon_path
        self.repaint()

    # DRAW ICON WITH COLORS
    # ///////////////////////////////////////////////////////////////
    def icon_paint(self, qp, image, rect, color):
        icon = QPixmap(image)
        painter = QPainter(icon)
        painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
        painter.fillRect(icon.rect(), color)
        qp.drawPixmap(
            (rect.width() - icon.width()) / 2, 
            (rect.height() - icon.height()) / 2,
            icon
        )        
        painter.end()

    # DRAW ACTIVE ICON / RIGHT SIDE
    # ///////////////////////////////////////////////////////////////
    def icon_active(self, qp, image, width):
        icon = QPixmap(image)
        painter = QPainter(icon)
        painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
        painter.fillRect(icon.rect(), self._bg_one)
        qp.drawPixmap(width - 5, 0, icon)
        painter.end()

    # CHANGE STYLES
    # Functions with custom styles
    # ///////////////////////////////////////////////////////////////
    def change_style(self, event):
        if event == QEvent.Enter:
            if not self._is_active:
                self._set_icon_color = self._icon_color_hover
                self._set_bg_color = self._dark_three
            self.repaint()          
        elif event == QEvent.Leave:
            if not self._is_active:
                self._set_icon_color = self._icon_color
                self._set_bg_color = self._dark_one
            self.repaint()
        elif event == QEvent.MouseButtonPress:
            if not self._is_active:         
                self._set_icon_color = self._context_color
                self._set_bg_color = self._dark_four
            self.repaint()  
        elif event == QEvent.MouseButtonRelease:
            if not self._is_active:
                self._set_icon_color = self._icon_color_hover
                self._set_bg_color = self._dark_three
            self.repaint()
    
    # MOUSE OVER
    # Event triggered when the mouse is over the BTN
    # ///////////////////////////////////////////////////////////////
    def enterEvent(self, event):
        self.change_style(QEvent.Enter)
        if self.width() == 50 and self._tooltip_text:
            self.move_tooltip()
            self.tooltip.show()

    # MOUSE LEAVE
    # Event fired when the mouse leaves the BTN
    # ///////////////////////////////////////////////////////////////
    def leaveEvent(self, event):
        self.change_style(QEvent.Leave)
        self.tooltip.hide()

    # MOUSE PRESS
    # Event triggered when the left button is pressed
    # ///////////////////////////////////////////////////////////////
    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.change_style(QEvent.MouseButtonPress)
            self.tooltip.hide()
            return self.clicked.emit()

    # MOUSE RELEASED
    # Event triggered after the mouse button is released
    # ///////////////////////////////////////////////////////////////
    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.change_style(QEvent.MouseButtonRelease)
            return self.released.emit()

    # MOVE TOOLTIP
    # ///////////////////////////////////////////////////////////////
    def move_tooltip(self):
        # GET MAIN WINDOW PARENT
        gp = self.mapToGlobal(QPoint(0, 0))

        # SET WIDGET TO GET POSTION
        # Return absolute position of widget inside app
        pos = self._parent.mapFromGlobal(gp)

        # FORMAT POSITION
        # Adjust tooltip position with offset
        pos_x = pos.x() + self.width() + 5
        pos_y = pos.y() + (self.width() - self.tooltip.height()) // 2

        # SET POSITION TO WIDGET
        # Move tooltip position
        self.tooltip.move(pos_x, pos_y) 

class _ToolTip(QLabel):
    # TOOLTIP / LABEL StyleSheet
    style_tooltip = """ 
    QLabel {{		
        background-color: {_dark_one};	
        color: {_text_foreground};
        padding-left: 10px;
        padding-right: 10px;
        border-radius: 17px;
        border: 0px solid transparent;
        border-left: 3px solid {_context_color};
        font: 800 9pt "Segoe UI";
    }}
    """

    def __init__(
        self,
        parent, 
        tooltip,
        dark_one,
        context_color,
        text_foreground
    ):
        QLabel.__init__(self)

        # LABEL SETUP
        style = self.style_tooltip.format(
            _dark_one = dark_one,
            _context_color = context_color,
            _text_foreground = text_foreground
        )
        self.setObjectName(u"label_tooltip")
        self.setStyleSheet(style)
        self.setMinimumHeight(34)
        self.setParent(parent)
        self.setText(tooltip)
        self.adjustSize()

        # SET DROP SHADOW
        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(30)
        self.shadow.setXOffset(0)
        self.shadow.setYOffset(0)
        self.shadow.setColor(QColor(0, 0, 0, 80))
        self.setGraphicsEffect(self.shadow)

    



# %% 结论：两段直线没有意义，两个斜率对各次谐波的影响是同步的
from pylab import np, plt
plt.style.use('ggplot')

V_AMPL = 1.0
I_AMPL = 10.0
PERIOD = 0.1 # s
CL_TS = 1e-4
time = np.arange(0, PERIOD, CL_TS)
phase_angle = np.arange(-np.pi, np.pi, 2*np.pi / len(time))
phase_current = np.sin(phase_angle) * I_AMPL

""" By Phase Angle """
def trapezoidal(theta_A, theta_t, V_plateau):

    oneOver_theta_t = 1.0 / theta_t

    if theta_A > 0:
        if theta_A < theta_t:
            compensation = theta_A * V_plateau * oneOver_theta_t
        elif theta_A > np.pi - theta_t:
            compensation = (np.pi-theta_A) * V_plateau * oneOver_theta_t
        else:
            compensation = V_plateau
    else:
        if -theta_A < theta_t:
            compensation = -theta_A * -V_plateau * oneOver_theta_t
        elif -theta_A > np.pi - theta_t:
            compensation = (np.pi+theta_A) * -V_plateau * oneOver_theta_t
        else:
            compensation = -V_plateau
    return compensation

""" By Current Value """
def tratrapezoizoidal(current, 
        R_0=12.0, V_plateau=V_AMPL, 
        R_1=3.5, V_turning=V_AMPL*0.7, 
        I_turning=V_AMPL*0.7/12.0, I_plateau=V_AMPL*0.7/12.0 + V_AMPL*(1-0.7) / 3.5):
    # R_1 must be less than R_0
    # V_turning = \kappa * V_plateau
    # I_turning = V_turning / R_0
    # I_plateau = I_turning + (V_plateau - V_turning) / R_1

    if current > I_plateau:
        return V_plateau
    elif current < -I_plateau:
        return -V_plateau
    elif current > I_turning:
        return V_turning + R_1*(current - I_turning)
    elif current < -I_turning:
        return -V_turning + R_1*(current + I_turning)
    else:
        return current * R_0

fig_UIcurve = plt.figure(figsize=(20,6))

def sigmoid(x):
    a1 = 0
    a2 = I_AMPL*2
    a3 = 15
    return a1 * x + a2 / (1.0 + np.exp(-a3 * x)) - a2*0.5

for theta_t in [ 5.0 / 180 * np.pi,
                # 10.0 / 180 * np.pi,
                # 15.0 / 180 * np.pi,
                # 20.0 / 180 * np.pi,
                ]:

    uic_t = [trapezoidal(theta_A, theta_t, I_AMPL) for theta_A in phase_angle]
    fig_UIcurve.gca().plot(phase_current, uic_t)

uic_s = [sigmoid(current_value) for current_value in phase_current]
fig_UIcurve.gca().plot(phase_current, uic_s)

uic_tt = [tratrapezoizoidal(current_value) for current_value in phase_current]
fig_UIcurve.gca().plot(phase_current, uic_tt)

fig_timeDomain = plt.figure(figsize=(20,6))
fig_timeDomain.gca().plot(time, uic_t)
fig_timeDomain.gca().plot(time, uic_s)
fig_timeDomain.gca().plot(time, uic_tt)
fig_timeDomain.gca().plot(time, phase_current)

# %%
import sys
sys.path.insert(1, r'D:\DrH\Codes\visualize')
from DFT_CJH import myDFT_single_channel, DFT_CJH

ax = plt.figure(figsize=(20,6)).gca()

REPEAT = 10
time = np.arange(0, REPEAT*PERIOD, CL_TS)

signal = np.array(REPEAT*list(uic_t))
DFT_CJH(ax, time, signal, bool_use_dB=False)
signal = np.array(REPEAT*list(uic_tt))
DFT_CJH(ax, time, signal, bool_use_dB=False)
signal = np.array(REPEAT*list(uic_s))
DFT_CJH(ax, time, signal, bool_use_dB=False)

ax.set_xlim([0,200])
# ax.set_ylim([-50,5])

print(len(time), len(signal))

# %%
REPEAT = 1
time = np.arange(0, REPEAT*PERIOD, CL_TS)

# Free parameters
OPTION = 1
if OPTION == 1:
    kappa = 0.7
    R_0 = 12.0 # 另一种思路是给定 I_plateau，这样R_0就相应被确定了。
    R_1 = 3.5
    V_plateau = I_AMPL # = 1
elif OPTION == 2:
    V_plateau = I_AMPL # = 1
    kappa = 0.7
    I_plateau = 0.15 # 0.25 # [A]
    R_1 = 4.5
elif OPTION == 3:
    V_plateau = I_AMPL # = 1
    kappa = 0.7
    I_plateau = 0.15 # 0.25 # [A]
    R_0 = 12.0 # [Ohm]

harmonics = dict()
harmonics[1] = []
harmonics[5] = []
harmonics[7] = []
harmonics[11] = []
harmonics[13] = []
if OPTION == 1: # I_plateau is dependent parameter
    LIST_R_0 = np.arange(1.0, 10, 0.25)
    LIST_R_1 = np.arange(0.1, R_0, 0.1)
    LIST_R_1_RATIO = np.arange(0.2, 1.1, 0.2)
elif OPTION == 2: # R_0 is dependent parameter
    LIST_R_1 = np.arange(2.5, 5.5, 0.2)
    LIST_I_plateau = np.arange(0.1, 0.6, 0.02)
elif OPTION == 3: # R_1 is dependent parameter
    LIST_R_0 = np.arange(0.5, 20, 0.5)
    LIST_I_plateau = np.arange(0.1, 0.6, 0.02)
LIST_kappa = np.arange(0.5, 0.95, 0.1)
fig_UIcurve = plt.figure(figsize=(20,6))
fig_timeDomain = plt.figure(figsize=(20,6))
LIST_I_plateau_plot = []

# LIST_PARAM = LIST_kappa
# for kappa in LIST_kappa:

# LIST_PARAM = LIST_I_plateau
# for I_plateau in LIST_PARAM:

# LIST_PARAM = LIST_R_1
# for R_1 in LIST_R_1:

# LIST_PARAM = LIST_R_0
# for R_0 in LIST_R_0:

# LIST_PARAM = LIST_R_1_RATIO
# for R_1_RATIO in LIST_R_1_RATIO:

for R_0 in [2, 5,  10, 100]:
    # for R_1_RATIO in [0.33, 1.0]:
    # for R_1_RATIO in [1.0, 1.5]:
    for R_1_RATIO in [1.0]:
        for kappa in [0.7]:
        #for kappa in [0.7]:
 
            # Dependent parameters
            V_turning = V_plateau*kappa
            if OPTION == 1:
                R_1 = R_1_RATIO * R_0
                I_plateau=V_turning/R_0 + (V_plateau-V_turning) / R_1
                I_turning = V_turning/R_0
            elif OPTION ==2:
                I_turning = I_plateau - (V_plateau-V_turning) / R_1
                R_0 = V_turning / I_turning

                print('R_0 =', R_0)
                if R_0<R_1:
                    R_0 = R_1 # become trapezoidal 
                    print(f'Put R_0=R_1={R_1}')
                    # continue

            elif OPTION ==3:
                I_turning = V_turning / R_0
                R_1 = V_turning / I_turning

                print('R_1, R_0 =', R_1, R_0)
                if R_0<R_1:
                    R_1 = R_0 # become trapezoidal 
                    print(f'Put R_1=R_0={R_0}')
                    # continue


            # uic_s = [sigmoid(current_value) for current_value in phase_current]
            # uic_t = [trapezoidal(theta_A, theta_t, AMPL) for theta_A in phase_angle]
            uic_tt = [tratrapezoizoidal(current_value, 
                            R_0=R_0, V_plateau=V_plateau,
                            R_1=R_1, V_turning=V_turning,
                            I_turning=I_turning, I_plateau=I_plateau,
                        ) for current_value in phase_current]

            fig_UIcurve.gca().plot(phase_current, uic_tt)
            fig_timeDomain.gca().plot(time, uic_tt)

            # signal = np.array(REPEAT*list(uic_s))
            # signal = np.array(REPEAT*list(uic_t))
            signal = np.array(REPEAT*list(uic_tt))

            dft_dict = myDFT_single_channel(time, signal)

            # print([el for el in dft_dict['ampl'][0:15]])
            # print([el for el in dft_dict['ampl'][0:15] if el>1e-5])

            if REPEAT == 1: # 如果REPEAT变了分辨率就变了，间隔也就变了哦。
                LIST_I_plateau_plot.append(I_plateau)
                harmonics[ 1].append(dft_dict['ampl'][ 1]); #print(' 10 Hz', dft_dict['ampl'][ 1])
                harmonics[ 5].append(dft_dict['ampl'][ 5]); #print(' 50 Hz', dft_dict['ampl'][ 5])
                harmonics[ 7].append(dft_dict['ampl'][ 7]); #print(' 70 Hz', dft_dict['ampl'][ 7])
                harmonics[11].append(dft_dict['ampl'][11]); #print('110 Hz', dft_dict['ampl'][11])
                harmonics[13].append(dft_dict['ampl'][13]); #print('130 Hz', dft_dict['ampl'][13])

# %%
# LIST_PARAM
# LIST_PARAM
# LIST_PARAM
# LIST_PARAM
# LIST_PARAM
plt.figure(figsize=(20,6))
plt.plot(harmonics[ 1], '.-')
plt.figure(figsize=(20,6))
plt.plot(harmonics[ 5], '.-')
plt.plot(harmonics[ 7], '.-')
plt.plot(harmonics[11], '.-')
plt.plot(harmonics[13], '.-')
plt.show()


# %%
from pylab import np, plt
def sigmoid(x, a1=0, a2=5, a3=10):
    return a1 * x + a2 / (1.0 + np.exp(-a3 * x)) - a2*0.5

x = np.arange(-10,10,1e-3)
for a3 in [1,3,5,7,9,100]:
    plt.plot(x, sigmoid(x, a3=a3))

# %%



# 常用库
from pylab import np, mpl, plt
from collections import OrderedDict as OD
from dataclasses import dataclass
import matplotlib.animation as animation
import json, re, copy, pkg_resources
import traceback
import pandas as pd

# 后端
# use cairo only for acmsimpy | use cairo for acmsimc will slow down plotting
# mpl.use('cairo') # try if this is faster for animate plotting? ['GTK3Agg', 'GTK3Cairo', 'MacOSX', 'nbAgg', 'Qt4Agg', 'Qt4Cairo', 'Qt5Agg', 'Qt5Cairo', 'TkAgg', 'TkCairo', 'WebAgg', 'WX', 'WXAgg', 'WXCairo', 'agg', 'cairo', 'pdf', 'pgf', 'ps', 'svg', 'template']

# 风格
mpl.style.use('ggplot') # further customize: https://stackoverflow.com/questions/35223312/matplotlib-overriding-ggplot-default-style-properties and https://matplotlib.org/2.0.2/users/customizing.html
# mpl.style.use("dark_background")
# mpl.style.use('grayscale')
# mpl.style.use('classic')

# 字体
mpl.rcParams['mathtext.fontset'] = 'stix'
mpl.rcParams['font.family'] = 'STIXGeneral'
mpl.rcParams['font.weight'] ='900'     # bold is 700, normal is 400, see https://matplotlib.org/2.0.2/users/customizing.html
mpl.rcParams['font.size']   =12 # 13 is too big, commenting out is too small

# 颜色设置壹/叁
# https://matplotlib.org/2.0.2/users/customizing.html from https://stackoverflow.com/questions/35223312/matplotlib-overriding-ggplot-default-style-properties
# mpl.rcParams['axes.labelsize']  =14   # 'medium' # 'large'是没用的！
hex_monokai   = '#272822'
hex_wanderson = '#3d444c'
mpl.rcParams['axes.facecolor']  =hex_wanderson # also need to change facecolor in mplwidget.py to get a consistent look
mpl.rcParams['axes.labelcolor'] ='#d5d1c7' # tint grey
mpl.rcParams['axes.axisbelow']  ='False'   # 'line'
mpl.rcParams['ytick.color']     ='#d5d1c7' #'white'
mpl.rcParams['xtick.color']     ='#d5d1c7' #'white'
mpl.rcParams['text.color']      ='#d5d1c7' #'white'
mpl.rcParams['grid.color']      ='#d5d1c7' # grid color
mpl.rcParams['grid.linestyle']  ='--'      # solid
mpl.rcParams['grid.linewidth']  =0.3       # in points
mpl.rcParams['grid.alpha']      =0.8       # transparency, between 0.0 and 1.0



@dataclass
class CJH_Plotting(object):
    cjh_linestyles = [
            '-', '--', (0, (3, 1, 1, 1)), ':', '-.', 
            '-', '--', (0, (3, 1, 1, 1)), ':', '-.', 
            '-', '--', (0, (3, 1, 1, 1)), ':', '-.', 
        ]
    # 颜色设置叁/叁
    cjh_colors = [
            '#ff6347', # tomato red
            'white',   # #d5d1c7',
            '#B3F5FF', # dark-theme-bright-blue
            '#FFF0B3', # dark-theme-bright-yellow
            '#ffabff', # pink
            '#E1C7E0', # dark-theme-bright-violet(purple)
            '#ABF5D1', # dark-theme-bright-green
            '#FFBDAD', # dark-theme-bright-red
            '#00dcff', # blue
            '#c593ff', # dark-theme-bright-color1
            '#02dac3', # dark-theme-bright-color2
            '#efc9a1', # dark-theme-bright-color3
            # 'black',
            'blue',
            '#857F72', # warmGrey
            '#616E7C', # Cool Grey
            '#B8B2A7', # warmGrey
            '#9AA5B1', # Cool Grey
            '#504A40', # warmGrey
            '#27241D', # warmGrey
        ] # https://uxdesign.cc/dark-mode-ui-design-the-definitive-guide-part-1-color-53dcfaea5129

    def __init__(self):
        # 作图用外观设置
        # generator的好处是循环，列表则是有限的 # https://matplotlib.org/3.1.0/gallery/lines_bars_and_markers/linestyles.html
        self.cjh_linestyle = self.cyclic_generator(self.cjh_linestyles)
        self.cjh_color = self.cyclic_generator(self.cjh_colors)
        # 注意，不可以做 list(self.cjh_color)，因为这是一个无止境循环的发生器，会卡住的。。。

    @staticmethod
    def cyclic_generator(list_of_things):
        N, i = len(list_of_things), 0
        while True:
            yield list_of_things[i]
            i +=1
            if i > (N-1):
                i = 0


from emachinery.acmsimpyv1 import acmsimpy

# FUNCTIONS
class EmyFunctions(object):

    # Run real time simulation with ACMSymPy
    ''' Python-Numba-based Simulation '''
    # @Slot()
    # @decorator_status(status='Configuring') # this will not be updated until this function is done
    def runPyBasedSimulation(self, bool_realtime=True):

        # 颜色设置贰/叁
        self.ui.load_pages.MplWidget_ACMPlot.toolbar.setStyleSheet("background-color: #9AA5B1;")
        # self.ui.right_column.horizontalLayout_CBSMplToolBar.addWidget(self.ui.MplWidget_ACMPlot.toolbar)

        def update_line_data(end_time, line, ydata):
            line.ydata = np.append(line.ydata, ydata)
            line.ydata = line.ydata[-CONSOLE.NUMBER_OF_SAMPLE_TO_SHOW:]
            line.MIN = min(line.ydata) if min(line.ydata) < line.MIN else line.MIN
            line.MAX = max(line.ydata) if max(line.ydata) > line.MAX else line.MAX

            xdata = np.arange(end_time*CONSOLE.SAMPLING_RATE-len(line.ydata), end_time*CONSOLE.SAMPLING_RATE, 1) * CONSOLE.CL_TS # 整数代替浮点数，精度高否则会出现xdata长3001，ydata长3000的问题。
            # print(len(xdata),len(line.ydata))
            line.set_data(xdata, line.ydata)
            if line.MIN != line.MAX:
                line.ax.set_ylim([line.MIN, line.MAX]) # this .ax is manually assigned when line object is created.

        def ACMアニメ4Py(ii):
            # if ii>100000 or self.ui.load_pages.lineEdit_AtTop.text() == 's':
            #     CONSOLE.anim.event_source.stop()
            #     print('Stop ACMPlot animation for Python.\n----------\n\n')
            #     UIFunctions.labelTitle(self, 'Main Window - emachinery - Simulation Done')

                # Save time here
                # print(f'Animation goes {ii}')
                # try:
                #     int(self.ui.load_pages.lineEdit_AtTop.text())
                # except:
                #     pass
                # else:
                #     self.NUMBER_OF_SAMPLE_TO_SHOW = int(self.ui.load_pages.lineEdit_AtTop.text())
                # if self.NUMBER_OF_SAMPLE_TO_SHOW<self.TIME_SLICE*self.SAMPLING_RATE:
                #     self.NUMBER_OF_SAMPLE_TO_SHOW = int(self.TIME_SLICE*self.SAMPLING_RATE)

                # Runtime User Input from window self.eW
                # if self.eW is not None and self.eW.input1.text() != '':
                #     CTRL.cmd_rpm_speed = float(self.eW.input1.text())
                #     print('\t', CTRL.cmd_rpm_speed, 'rpm')

            """perform animation step"""
            if CONSOLE.reset == True:
                CONSOLE.reset = False
                CONSOLE.offset_anime_ii = ii
            t0 = (ii-CONSOLE.offset_anime_ii)*CONSOLE.TIME_SLICE
            control_times, numba__waveforms_dict, = \
                acmsimpy.ACMSimPyWrapper(
                    numba__scope_dict,
                    t0=t0, TIME=CONSOLE.TIME_SLICE,
                    ACM=ACM,
                    CTRL=CTRL,
                    reg_id=reg_id,
                    reg_iq=reg_iq,
                    reg_speed=reg_speed,
                    OB=OB,
                    HUMAN=HUMAN
                )
            end_time = control_times[-1] # print('\tend_time:', end_time)

            """ update matplotlib artist """
            for key, waveforms_data in numba__waveforms_dict.items():
                for line, ydata in zip(numba__line_dict[key], waveforms_data):
                    # print(key, line, ydata[0:3], numba__line_dict[key], len(waveforms_data))
                    update_line_data(end_time, line, ydata)
            # time_text.set_text('time = %.1f' % end_time)

            first_ax.set_xlim([end_time-CONSOLE.NUMBER_OF_SAMPLE_TO_SHOW*CONSOLE.CL_TS, end_time])
            # first_ax.set_xticklabels(np.arange(0, int(round(end_time)), int(round(end_time))*0.1)) #, fontdict=font)

            # return line01, line02, time_text # for using blit=True but blit does not update xticks and yticks

        """ Read from GUI """
        try:
            the_cmd = self.ui.load_pages.plainTextEdit_NumbaScopeDict.toPlainText()
            numba__scope_dict = eval(the_cmd[the_cmd.find('OD'):])
        except Exception as err:
            print('-------------------')
            print('Decypher for NumbaScopeDict has failed. Will use the default dict.')
            numba__scope_dict = OD([
                (r'Speed [rpm]',                  ( 'CTRL.omega_elec'           ,)  ),
                (r'Position [rad]',               ( 'ACM.x[3]'                  ,)  ),
                (r'$q$-axis current [A]',         ( 'ACM.x[1]', 'CTRL.idq[1]'   ,)  ),
                (r'$d$-axis current [A]',         ( 'ACM.x[0]', 'CTRL.idq[0]'   ,)  ),
                (r'$\alpha\beta$ current [A]',    ( 'CTRL.iab[0]', 'CTRL.iab[1]',)  ),
            ])
            print('err:', err)
            traceback.print_exc()
            print('-------------------')

        """ Canvas """
        self.ui.load_pages.MplWidget_ACMPlot.canvas.figure.clf() # clear canvas
        print('numba__scope_dict =', numba__scope_dict)
        number_of_subplot = len(numba__scope_dict)
        numba__line_dict = OD()
        numba__axes = []
        first_ax = None

        cjh_plotting = CJH_Plotting()

        for index, (ylabel, waveform_names) in enumerate(numba__scope_dict.items()):
            # get axes
            ax = self.ui.load_pages.MplWidget_ACMPlot.canvas.figure.add_subplot(
                number_of_subplot, 
                1, 1+index, 
                autoscale_on=False, 
                sharex=first_ax
            )
            first_ax = ax if index == 0 else first_ax
            numba__axes.append(ax)

            numba__line_dict[ylabel] = []
            for jj, name in enumerate(waveform_names):
                # print(jj, name)
                # get line objects
                line, = ax.plot([], [], 
                                linestyle=cjh_plotting.cjh_linestyles[jj], 
                                color=cjh_plotting.cjh_colors[jj], 
                                lw=1.5, 
                                label=name, 
                                alpha=0.7) # zorder
                line.ax = ax
                line.ydata = np.array(0)
                line.MIN = -1e-10
                line.MAX =  1e-10
                numba__line_dict[ylabel].append(line)

            ax.set_ylabel(ylabel)
            ax.legend(loc='lower left').set_zorder(202)

        numba__axes[-1].set_xlabel('Time [s]')
        # time_text = first_ax.text(0.02, 0.95, '', transform=first_ax.transAxes)

        """ User control over the mpl animation """
        @dataclass
        class THE_CONSOLE:
            offset_anime_ii : int = 0
            reset : int = False
            _pause : int = False
            SAMPLING_RATE : float = 1e4
            TIME_SLICE : float = 0.2
            NUMBER_OF_SAMPLE_TO_SHOW : int = 30000
            def __post_init__(self):
                self.CL_TS = 1 / self.SAMPLING_RATE
        CONSOLE = THE_CONSOLE(TIME_SLICE=0.2)

        """ Simulation Globals """
        HUMAN     = acmsimpy.The_Human()
        OB        = acmsimpy.The_Observer(HUMAN)
        ACM       = acmsimpy.The_AC_Machines(HUMAN)
        CTRL      = acmsimpy.The_Motor_Controller(CONSOLE.CL_TS, 5*CONSOLE.CL_TS, HUMAN)
        reg_id    = acmsimpy.The_PI_Regulator(6.39955, 6.39955*237.845*CTRL.CL_TS, 600)
        reg_iq    = acmsimpy.The_PI_Regulator(6.39955, 6.39955*237.845*CTRL.CL_TS, 600)
        reg_speed = acmsimpy.The_PI_Regulator(0.0380362, 0.0380362*30.5565*CTRL.VL_TS, 1*1.414*ACM.IN)

        """ Simulation Globals Access from Console """
        self.console_push_variable({'CONSOLE':CONSOLE})
        self.console_push_variable({'ACM':ACM})
        self.console_push_variable({'CTRL':CTRL})
        self.console_push_variable({'HM':HUMAN, 'HUMAN':HUMAN})
        self.console_push_variable({'reg_id':reg_id})
        self.console_push_variable({'reg_iq':reg_iq})
        self.console_push_variable({'reg_speed':reg_speed})

        """ Visualization of Realtime Simulation """
        print('\tJIT compile with numba...')
        def onClick(event):
            print(event)
            CONSOLE._pause ^= True
            if CONSOLE._pause:
                CONSOLE.anim.event_source.stop()
            else:
                CONSOLE.anim.event_source.start()

        self.ui.load_pages.MplWidget_ACMPlot.canvas.figure.canvas.mpl_connect('button_press_event', onClick)
        self.ui.load_pages.MplWidget_ACMPlot.canvas.draw_idle() # draw before animation # https://stackoverflow.com/questions/8955869/why-is-plotting-with-matplotlib-so-slow
        CONSOLE.anim = animation.FuncAnimation(self.ui.load_pages.MplWidget_ACMPlot.canvas.figure, ACMアニメ4Py, 
                                        interval=200, cache_frame_data=False) # cache_frame_data=False is good for large artists when plotting
                                        #, blit=True) # blit=True means only re-draw the parts that have changed. 
                                        # https://matplotlib.org/stable/tutorials/advanced/blitting.html
                                        # https://stackoverflow.com/questions/14421924/matplotlib-animate-does-not-update-tick-labels
                                        # https://jakevdp.github.io/blog/2012/08/18/matplotlib-animation-tutorial/




# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT QT CORE
# ///////////////////////////////////////////////////////////////
from qt_core import *

# PY ICON WITH CUSTOM COLORS
# ///////////////////////////////////////////////////////////////
class PyIcon(QWidget):
    def __init__(
        self,
        icon_path,
        icon_color
    ):
        super().__init__()

        # PROPERTIES
        self._icon_path = icon_path
        self._icon_color = icon_color

        # SETUP UI
        self.setup_ui()

    def setup_ui(self):
        # LAYOUT
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0,0,0,0)

        # LABEL
        self.icon = QLabel()
        self.icon.setAlignment(Qt.AlignCenter)
        
        # PAINTER
        self.set_icon(self._icon_path, self._icon_color)

        # ADD TO LAYOUT
        self.layout.addWidget(self.icon)

    def set_icon(self, icon_path, icon_color = None):
        # GET COLOR
        color = ""
        if icon_color != None:
            color = icon_color
        else:
            color = self._icon_color

        # PAINTER / PIXMAP
        icon = QPixmap(icon_path)
        painter = QPainter(icon)
        painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
        painter.fillRect(icon.rect(), color)       
        painter.end()

        # SET PIXMAP
        self.icon.setPixmap(icon)

